import pytest

pytestmark = [
    pytest.mark.slow_test,
]


def test_orchestrate_output(salt_run_cli, salt_minion, salt_master):
    """
    Ensure the orchestrate runner outputs useful state data.

    In Issue #31330, the output only contains ['outputter:', '    highstate'],
    and not the full stateful return. This tests ensures we don't regress in that
    manner again.

    Also test against some sample "good" output that would be included in a correct
    orchestrate run.
    """
    bad_out = ["outputter:", "    highstate"]
    good_out = [
        "    Function: salt.state",
        "      Result: True",
        "Succeeded: 1 (changed=1)",
        "Failed:    0",
        "Total states run:     1",
    ]
    sls_contents = """
    call_sleep_state:
      salt.state:
        - tgt: {}
        - sls: simple-ping
    """.format(
        salt_minion.id
    )
    simple_ping_sls = """
    simple-ping:
      module.run:
        - name: test.ping
    """
    with salt_master.state_tree.base.temp_file(
        "orch-test.sls", sls_contents
    ), salt_master.state_tree.base.temp_file("simple-ping.sls", simple_ping_sls):
        ret = salt_run_cli.run("--out=highstate", "state.orchestrate", "orch-test")
        assert ret.returncode == 0
        ret_output = ret.stdout.splitlines()

        # First, check that we don't have the "bad" output that was displaying in
        # Issue #31330 where only the highstate outputter was listed
        assert bad_out != ret_output
        assert len(ret_output) > 2

        # Now test that some expected good sample output is present in the return.
        for item in good_out:
            assert item in ret_output


def test_orchestrate_state_output_with_salt_function(
    salt_run_cli, salt_minion, salt_master
):
    """
    Ensure that orchestration produces the correct output with salt.function.

    A salt execution module function does not return highstate data, so we
    should not try to recursively output it as such.
    The outlier to this rule is state.apply, but that is handled by the salt.state.

    See https://github.com/saltstack/salt/issues/60029 for more detail.
    """
    sls_contents = """
    arg_clean_test:
      salt.function:
        - name: test.arg_clean
        - arg:
          - B flat major
          - has 2 flats
        - tgt: {minion_id}

    ping_test:
      salt.function:
        - name: test.ping
        - tgt: {minion_id}
    """.format(
        minion_id=salt_minion.id
    )
    with salt_master.state_tree.base.temp_file("orch-function-test.sls", sls_contents):
        ret = salt_run_cli.run(
            "--out=highstate", "state.orchestrate", "orch-function-test"
        )
        assert ret.returncode == 0
        ret_output = [line.strip() for line in ret.stdout.splitlines()]

        assert "args:" in ret_output
        assert "- B flat major" in ret_output
        assert "- has 2 flats" in ret_output
        assert "True" in ret_output


def test_orchestrate_nested(salt_run_cli, salt_minion, salt_master, tmp_path):
    """
    test salt-run state.orchestrate and failhard with nested orchestration
    """
    testfile = tmp_path / "ewu-2016-12-13"
    inner_sls = """
    cmd.run:
      salt.function:
        - tgt: {}
        - arg:
          - {}
        - failhard: True
    """.format(
        salt_minion.id, pytest.helpers.shell_test_false()
    )
    outer_sls = """
    state.orchestrate:
      salt.runner:
        - mods: nested.inner
        - failhard: True

    cmd.run:
      salt.function:
        - tgt: {}
        - arg:
          - touch {}
    """.format(
        salt_minion.id, testfile
    )

    with salt_master.state_tree.base.temp_file(
        "nested/inner.sls", inner_sls
    ), salt_master.state_tree.base.temp_file("nested/outer.sls", outer_sls):
        ret = salt_run_cli.run("state.orchestrate", "nested.outer")
        assert ret.returncode != 0
        assert testfile.exists() is False


def test_orchestrate_with_mine(salt_run_cli, salt_minion, salt_master):
    """
    test salt-run state.orchestrate with mine.get call in sls
    """
    sls_contents = (
        """
    {% set minion = '"""
        + salt_minion.id
        + """' %}
    {% set mine = salt.saltutil.runner('mine.get', tgt=minion, fun='test.ping') %}

    {% if mine %}
    test.ping:
      salt.function:
        - tgt: "{{ minion }}"
    {% endif %}
    """
    )
    ret = salt_run_cli.run("mine.update", salt_minion.id)
    assert ret.returncode == 0

    with salt_master.state_tree.base.temp_file("orch/mine.sls", sls_contents):
        ret = salt_run_cli.run("state.orchestrate", "orch.mine")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data["data"][salt_master.id]
        for state_data in ret.data["data"][salt_master.id].values():
            assert state_data["changes"]["ret"]
            assert state_data["changes"]["ret"][salt_minion.id] is True


def test_orchestrate_state_and_function_failure(salt_run_cli, salt_master, salt_minion):
    """
    Ensure that returns from failed minions are in the changes dict where
    they belong, so they can be programmatically analyzed.

    See https://github.com/saltstack/salt/issues/43204
    """
    init_sls = """
    Step01:
      salt.state:
        - tgt: {minion_id}
        - sls:
          - orch.issue43204.fail_with_changes

    Step02:
      salt.function:
        - name: runtests_helpers.nonzero_retcode_return_false
        - tgt: {minion_id}
        - fail_function: runtests_helpers.fail_function
    """.format(
        minion_id=salt_minion.id
    )
    fail_sls = """
    test fail with changes:
      test.fail_with_changes
    """
    with salt_master.state_tree.base.temp_file(
        "orch/issue43204/init.sls", init_sls
    ), salt_master.state_tree.base.temp_file(
        "orch/issue43204/fail_with_changes.sls", fail_sls
    ):
        ret = salt_run_cli.run("saltutil.sync_modules")
        assert ret.returncode == 0

        ret = salt_run_cli.run("state.orchestrate", "orch.issue43204")
        assert ret.returncode != 0

    # Drill down to the changes dict
    data = ret.data["data"][salt_master.id]
    state_ret = data["salt_|-Step01_|-Step01_|-state"]["changes"]
    func_ret = data[
        "salt_|-Step02_|-runtests_helpers.nonzero_retcode_return_false_|-function"
    ]["changes"]

    # Remove duration and start time from the results, since they would
    # vary with each run and that would make it impossible to test.
    for item in ("duration", "start_time"):
        state_ret["ret"][salt_minion.id][
            "test_|-test fail with changes_|-test fail with changes_|-fail_with_changes"
        ].pop(item)

    expected = {
        "out": "highstate",
        "ret": {
            salt_minion.id: {
                "test_|-test fail with changes_|-test fail with changes_|-fail_with_changes": {
                    "__id__": "test fail with changes",
                    "__run_num__": 0,
                    "__sls__": "orch.issue43204.fail_with_changes",
                    "changes": {
                        "testing": {
                            "new": "Something pretended to change",
                            "old": "Unchanged",
                        }
                    },
                    "comment": "Failure!",
                    "name": "test fail with changes",
                    "result": False,
                }
            }
        },
    }
    assert state_ret == expected
    assert func_ret == {"ret": {salt_minion.id: False}}


def test_orchestrate_salt_function_return_false_failure(
    salt_run_cli, salt_minion, salt_master
):
    """
    Ensure that functions that only return False in the return
    are flagged as failed when run as orchestrations.

    See https://github.com/saltstack/salt/issues/30367
    """
    sls_contents = """
    deploy_check:
      salt.function:
        - name: test.false
        - tgt: {}
    """.format(
        salt_minion.id
    )
    with salt_master.state_tree.base.temp_file("orch/issue30367.sls", sls_contents):
        ret = salt_run_cli.run("saltutil.sync_modules")
        assert ret.returncode == 0

        ret = salt_run_cli.run("state.orchestrate", "orch.issue30367")
        assert ret.returncode != 0

    # Drill down to the changes dict
    data = ret.data["data"][salt_master.id]
    state_result = data["salt_|-deploy_check_|-test.false_|-function"]["result"]
    func_ret = data["salt_|-deploy_check_|-test.false_|-function"]["changes"]

    assert state_result is False
    assert func_ret == {"ret": {salt_minion.id: False}}


def test_orchestrate_target_exists(salt_run_cli, salt_minion, salt_master):
    """
    test orchestration when target exists while using multiple states
    """
    sls_contents = """
    core:
      salt.state:
        - tgt: '{minion_id}*'
        - sls:
          - core

    test-state:
      salt.state:
        - tgt: '{minion_id}*'
        - sls:
          - orch.target-test

    cmd.run:
      salt.function:
        - tgt: '{minion_id}*'
        - arg:
          - echo test
    """.format(
        minion_id=salt_minion.id
    )
    target_test_sls = """
    always_true:
      test.succeed_without_changes
    """
    with salt_master.state_tree.base.temp_file(
        "orch/target-exists.sls", sls_contents
    ), salt_master.state_tree.base.temp_file(
        "orch/target-test.sls", target_test_sls
    ), salt_master.state_tree.base.temp_file(
        "core.sls", target_test_sls
    ):
        ret = salt_run_cli.run("state.orchestrate", "orch.target-exists")
        assert ret.returncode == 0
        assert ret.data

    data = ret.data["data"][salt_master.id]
    to_check = {"core", "test-state", "cmd.run"}
    for state_data in data.values():
        if state_data["name"] == "core":
            to_check.remove("core")
            assert state_data["result"] is True
        if state_data["name"] == "test-state":
            assert state_data["result"] is True
            to_check.remove("test-state")
        if state_data["name"] == "cmd.run":
            assert state_data["changes"] == {
                "ret": {salt_minion.id: "test"},
            }
            to_check.remove("cmd.run")

    assert not to_check


def test_orchestrate_target_does_not_exist(salt_run_cli, salt_minion, salt_master):
    """
    test orchestration when target does not exist while using multiple states
    """
    sls_contents = """
    core:
      salt.state:
        - tgt: 'does-not-exist*'
        - sls:
          - core

    test-state:
      salt.state:
        - tgt: '{minion_id}*'
        - sls:
          - orch.target-test

    cmd.run:
      salt.function:
        - tgt: '{minion_id}*'
        - arg:
          - echo test
    """.format(
        minion_id=salt_minion.id
    )
    target_test_sls = """
    always_true:
      test.succeed_without_changes
    """
    with salt_master.state_tree.base.temp_file(
        "orch/target-does-not-exist.sls", sls_contents
    ), salt_master.state_tree.base.temp_file(
        "orch/target-test.sls", target_test_sls
    ), salt_master.state_tree.base.temp_file(
        "core.sls", target_test_sls
    ):
        ret = salt_run_cli.run("state.orchestrate", "orch.target-does-not-exist")
        assert ret.returncode != 0
        assert ret.data

    data = ret.data["data"][salt_master.id]
    to_check = {"core", "test-state", "cmd.run"}
    for state_data in data.values():
        if state_data["name"] == "core":
            to_check.remove("core")
            assert state_data["result"] is False
            assert state_data["comment"] == "No minions returned"
        if state_data["name"] == "test-state":
            assert state_data["result"] is True
            to_check.remove("test-state")
        if state_data["name"] == "cmd.run":
            assert state_data["changes"] == {
                "ret": {salt_minion.id: "test"},
            }
            to_check.remove("cmd.run")

    assert not to_check


def test_orchestrate_retcode(salt_run_cli, salt_master):
    """
    Test orchestration with nonzero retcode set in __context__
    """
    sls_contents = """
    test_runner_success:
      salt.runner:
        - name: runtests_helpers.success

    test_runner_failure:
      salt.runner:
        - name: runtests_helpers.failure

    test_wheel_success:
      salt.wheel:
        - name: runtests_helpers.success

    test_wheel_failure:
      salt.wheel:
        - name: runtests_helpers.failure
    """
    with salt_master.state_tree.base.temp_file("orch/retcode.sls", sls_contents):
        ret = salt_run_cli.run("saltutil.sync_runners")
        assert ret.returncode == 0
        ret = salt_run_cli.run("saltutil.sync_wheel")
        assert ret.returncode == 0

        ret = salt_run_cli.run("state.orchestrate", "orch.retcode")
        assert ret.returncode != 0
        assert ret.data

    data = ret.data["data"][salt_master.id]
    to_check = {
        "test_runner_success",
        "test_runner_failure",
        "test_wheel_failure",
        "test_wheel_success",
    }

    for state_data in data.values():
        name = state_data["__id__"]
        to_check.remove(name)
        if name in ("test_runner_success", "test_wheel_success"):
            assert state_data["result"] is True
        if name in ("test_runner_failure", "test_wheel_failure"):
            assert state_data["result"] is False

    assert not to_check


def test_orchestrate_batch_with_failhard_error(
    salt_run_cli, salt_master, salt_minion, tmp_path
):
    """
    test orchestration properly stops with failhard and batch.
    """
    testfile = tmp_path / "test-file"
    sls_contents = """
    call_fail_state:
      salt.state:
        - tgt: {}
        - batch: 1
        - failhard: True
        - sls: fail
    """.format(
        salt_minion.id
    )
    fail_sls = """
    {}:
      file.managed:
        - source: salt://hnlcfsdjhkzkdhynclarkhmcls
    """.format(
        testfile
    )
    with salt_master.state_tree.base.temp_file(
        "orch/batch.sls", sls_contents
    ), salt_master.state_tree.base.temp_file("fail.sls", fail_sls):
        ret = salt_run_cli.run("state.orchestrate", "orch.batch")
        assert ret.returncode != 0

    data = ret.data["data"][salt_master.id]
    result = data["salt_|-call_fail_state_|-call_fail_state_|-state"]["result"]
    changes = data["salt_|-call_fail_state_|-call_fail_state_|-state"]["changes"]

    assert result is False
    # The execution should stop after first error, so return dict should contain only one minion
    assert len(changes["ret"]) == 1


def _check_skip(grains):
    if grains["os"] == "Fedora":
        return True
    if grains["os"] == "VMware Photon OS" and grains["osmajorrelease"] == 4:
        return True
    if grains["os"] == "Ubuntu" and grains["osmajorrelease"] in (20, 22):
        return True
    return False


@pytest.mark.timeout_unless_on_windows(120)
@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_orchestrate_subset(
    salt_run_cli,
    salt_master,
    salt_minion,
    salt_sub_minion,
):
    """
    test orchestration state using subset
    """
    sls_contents = """
    test subset:
      salt.state:
        - tgt: '*minion*'
        - subset: 1
        - sls: test
    """
    test_sls = """
    test state:
      test.succeed_without_changes:
        - name: test
    """
    with salt_master.state_tree.base.temp_file(
        "orch/subset.sls", sls_contents
    ), salt_master.state_tree.base.temp_file("test.sls", test_sls):
        ret = salt_run_cli.run("state.orchestrate", "orch.subset", _timeout=60)
        assert ret.returncode == 0

    for state_data in ret.data["data"][salt_master.id].values():
        # Should only run in one of the minions
        comment = state_data["comment"]
        if salt_minion.id in comment:
            assert salt_sub_minion.id not in comment
        elif salt_sub_minion.id in comment:
            assert salt_minion.id not in comment
        else:
            pytest.fail(
                "None of the targeted minions({}) show up in comment: '{}'".format(
                    ", ".join([salt_minion.id, salt_sub_minion.id]), comment
                )
            )



"""
Manage Apache Sites

.. versionadded:: 2016.3.0

Enable and disable apache sites.

.. code-block:: yaml

    Enable default site:
      apache_site.enabled:
        - name: default

    Disable default site:
      apache_site.disabled:
        - name: default
"""


def __virtual__():
    """
    Only load if a2ensite is available.
    """
    if "apache.a2ensite" in __salt__:
        return "apache_site"
    return (False, "apache module could not be loaded")


def enabled(name):
    """
    Ensure an Apache site is enabled.

    name
        Name of the Apache site
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_site_enabled"](name)
    if not is_enabled:
        if __opts__["test"]:
            msg = f"Apache site {name} is set to be enabled."
            ret["comment"] = msg
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
            ret["result"] = None
            return ret
        status = __salt__["apache.a2ensite"](name)["Status"]
        if isinstance(status, str) and "enabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to enable {name} Apache site"
            if isinstance(status, str):
                ret["comment"] = ret["comment"] + f" ({status})"
            return ret
    else:
        ret["comment"] = f"{name} already enabled."
    return ret


def disabled(name):
    """
    Ensure an Apache site is disabled.

    name
        Name of the Apache site
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_site_enabled"](name)
    if is_enabled:
        if __opts__["test"]:
            msg = f"Apache site {name} is set to be disabled."
            ret["comment"] = msg
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
            ret["result"] = None
            return ret
        status = __salt__["apache.a2dissite"](name)["Status"]
        if isinstance(status, str) and "disabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to disable {name} Apache site"
            if isinstance(status, str):
                ret["comment"] = ret["comment"] + f" ({status})"
            return ret
    else:
        ret["comment"] = f"{name} already disabled."
    return ret



import ctypes
import multiprocessing

import pytest

import salt.crypt
import salt.master
import salt.utils.stringutils


@pytest.fixture(autouse=True)
def _prepare_aes():
    old_aes = salt.master.SMaster.secrets.get("aes")
    try:
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "reload": salt.crypt.Crypticle.generate_key_string,
        }
    finally:
        if old_aes:
            salt.master.SMaster.secrets["aes"] = old_aes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: skip-file


from __future__ import absolute_import, division, print_function
from salt.ext.tornado.httputil import url_concat, parse_multipart_form_data, HTTPHeaders, format_timestamp, HTTPServerRequest, parse_request_start_line, parse_cookie
from salt.ext.tornado.escape import utf8, native_str
from salt.ext.tornado.log import gen_log
from salt.ext.tornado.testing import ExpectLog
from salt.ext.tornado.test.util import unittest

import copy
import datetime
import logging
import pickle
import time


class TestUrlConcat(unittest.TestCase):
    def test_url_concat_no_query_params(self):
        url = url_concat(
            "https://localhost/path",
            [('y', 'y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?y=y&z=z")

    def test_url_concat_encode_args(self):
        url = url_concat(
            "https://localhost/path",
            [('y', '/y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?y=%2Fy&z=z")

    def test_url_concat_trailing_q(self):
        url = url_concat(
            "https://localhost/path?",
            [('y', 'y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?y=y&z=z")

    def test_url_concat_q_with_no_trailing_amp(self):
        url = url_concat(
            "https://localhost/path?x",
            [('y', 'y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?x=&y=y&z=z")

    def test_url_concat_trailing_amp(self):
        url = url_concat(
            "https://localhost/path?x&",
            [('y', 'y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?x=&y=y&z=z")

    def test_url_concat_mult_params(self):
        url = url_concat(
            "https://localhost/path?a=1&b=2",
            [('y', 'y'), ('z', 'z')],
        )
        self.assertEqual(url, "https://localhost/path?a=1&b=2&y=y&z=z")

    def test_url_concat_no_params(self):
        url = url_concat(
            "https://localhost/path?r=1&t=2",
            [],
        )
        self.assertEqual(url, "https://localhost/path?r=1&t=2")

    def test_url_concat_none_params(self):
        url = url_concat(
            "https://localhost/path?r=1&t=2",
            None,
        )
        self.assertEqual(url, "https://localhost/path?r=1&t=2")

    def test_url_concat_with_frag(self):
        url = url_concat(
            "https://localhost/path#tab",
            [('y', 'y')],
        )
        self.assertEqual(url, "https://localhost/path?y=y#tab")

    def test_url_concat_multi_same_params(self):
        url = url_concat(
            "https://localhost/path",
            [('y', 'y1'), ('y', 'y2')],
        )
        self.assertEqual(url, "https://localhost/path?y=y1&y=y2")

    def test_url_concat_multi_same_query_params(self):
        url = url_concat(
            "https://localhost/path?r=1&r=2",
            [('y', 'y')],
        )
        self.assertEqual(url, "https://localhost/path?r=1&r=2&y=y")

    def test_url_concat_dict_params(self):
        url = url_concat(
            "https://localhost/path",
            dict(y='y'),
        )
        self.assertEqual(url, "https://localhost/path?y=y")


class MultipartFormDataTest(unittest.TestCase):
    def test_file_upload(self):
        data = b"""\
--1234
Content-Disposition: form-data; name="files"; filename="ab.txt"

Foo
--1234--""".replace(b"\n", b"\r\n")
        args = {}
        files = {}
        parse_multipart_form_data(b"1234", data, args, files)
        file = files["files"][0]
        self.assertEqual(file["filename"], "ab.txt")
        self.assertEqual(file["body"], b"Foo")

    def test_unquoted_names(self):
        # quotes are optional unless special characters are present
        data = b"""\
--1234
Content-Disposition: form-data; name=files; filename=ab.txt

Foo
--1234--""".replace(b"\n", b"\r\n")
        args = {}
        files = {}
        parse_multipart_form_data(b"1234", data, args, files)
        file = files["files"][0]
        self.assertEqual(file["filename"], "ab.txt")
        self.assertEqual(file["body"], b"Foo")

    def test_special_filenames(self):
        filenames = ['a;b.txt',
                     'a"b.txt',
                     'a";b.txt',
                     'a;"b.txt',
                     'a";";.txt',
                     'a\\"b.txt',
                     'a\\b.txt',
                     ]
        for filename in filenames:
            logging.debug("trying filename %r", filename)
            data = """\
--1234
Content-Disposition: form-data; name="files"; filename="%s"

Foo
--1234--""" % filename.replace('\\', '\\\\').replace('"', '\\"')
            data = utf8(data.replace("\n", "\r\n"))
            args = {}
            files = {}
            parse_multipart_form_data(b"1234", data, args, files)
            file = files["files"][0]
            self.assertEqual(file["filename"], filename)
            self.assertEqual(file["body"], b"Foo")

    def test_boundary_starts_and_ends_with_quotes(self):
        data = b'''\
--1234
Content-Disposition: form-data; name="files"; filename="ab.txt"

Foo
--1234--'''.replace(b"\n", b"\r\n")
        args = {}
        files = {}
        parse_multipart_form_data(b'"1234"', data, args, files)
        file = files["files"][0]
        self.assertEqual(file["filename"], "ab.txt")
        self.assertEqual(file["body"], b"Foo")

    def test_missing_headers(self):
        data = b'''\
--1234

Foo
--1234--'''.replace(b"\n", b"\r\n")
        args = {}
        files = {}
        with ExpectLog(gen_log, "multipart/form-data missing headers"):
            parse_multipart_form_data(b"1234", data, args, files)
        self.assertEqual(files, {})

    def test_invalid_content_disposition(self):
        data = b'''\
--1234
Content-Disposition: invalid; name="files"; filename="ab.txt"

Foo
--1234--'''.replace(b"\n", b"\r\n")
        args = {}
        files = {}
        with ExpectLog(gen_log, "Invalid multipart/form-data"):
            parse_multipart_form_data(b"1234", data, args, files)
        self.assertEqual(files, {})

    def test_line_does_not_end_with_correct_line_break(self):
        data = b'''\
--1234
Content-Disposition: form-data; name="files"; filename="ab.txt"

Foo--1234--'''.replace(b"\n", b"\r\n")
        args = {}
        files = {}
        with ExpectLog(gen_log, "Invalid multipart/form-data"):
            parse_multipart_form_data(b"1234", data, args, files)
        self.assertEqual(files, {})

    def test_content_disposition_header_without_name_parameter(self):
        data = b"""\
--1234
Content-Disposition: form-data; filename="ab.txt"

Foo
--1234--""".replace(b"\n", b"\r\n")
        args = {}
        files = {}
        with ExpectLog(gen_log, "multipart/form-data value missing name"):
            parse_multipart_form_data(b"1234", data, args, files)
        self.assertEqual(files, {})

    def test_data_after_final_boundary(self):
        # The spec requires that data after the final boundary be ignored.
        # http://www.w3.org/Protocols/rfc1341/7_2_Multipart.html
        # In practice, some libraries include an extra CRLF after the boundary.
        data = b"""\
--1234
Content-Disposition: form-data; name="files"; filename="ab.txt"

Foo
--1234--
""".replace(b"\n", b"\r\n")
        args = {}
        files = {}
        parse_multipart_form_data(b"1234", data, args, files)
        file = files["files"][0]
        self.assertEqual(file["filename"], "ab.txt")
        self.assertEqual(file["body"], b"Foo")


class HTTPHeadersTest(unittest.TestCase):
    def test_multi_line(self):
        # Lines beginning with whitespace are appended to the previous line
        # with any leading whitespace replaced by a single space.
        # Note that while multi-line headers are a part of the HTTP spec,
        # their use is strongly discouraged.
        data = """\
Foo: bar
 baz
Asdf: qwer
\tzxcv
Foo: even
     more
     lines
""".replace("\n", "\r\n")
        headers = HTTPHeaders.parse(data)
        self.assertEqual(headers["asdf"], "qwer zxcv")
        self.assertEqual(headers.get_list("asdf"), ["qwer zxcv"])
        self.assertEqual(headers["Foo"], "bar baz,even more lines")
        self.assertEqual(headers.get_list("foo"), ["bar baz", "even more lines"])
        self.assertEqual(sorted(list(headers.get_all())),
                         [("Asdf", "qwer zxcv"),
                          ("Foo", "bar baz"),
                          ("Foo", "even more lines")])

    def test_unicode_newlines(self):
        # Ensure that only \r\n is recognized as a header separator, and not
        # the other newline-like unicode characters.
        # Characters that are likely to be problematic can be found in
        # http://unicode.org/standard/reports/tr13/tr13-5.html
        # and cpython's unicodeobject.c (which defines the implementation
        # of unicode_type.splitlines(), and uses a different list than TR13).
        newlines = [
            u'\u001b',  # VERTICAL TAB
            u'\u001c',  # FILE SEPARATOR
            u'\u001d',  # GROUP SEPARATOR
            u'\u001e',  # RECORD SEPARATOR
            u'\u0085',  # NEXT LINE
            u'\u2028',  # LINE SEPARATOR
            u'\u2029',  # PARAGRAPH SEPARATOR
        ]
        for newline in newlines:
            # Try the utf8 and latin1 representations of each newline
            for encoding in ['utf8', 'latin1']:
                try:
                    try:
                        encoded = newline.encode(encoding)
                    except UnicodeEncodeError:
                        # Some chars cannot be represented in latin1
                        continue
                    data = b'Cookie: foo=' + encoded + b'bar'
                    # parse() wants a native_str, so decode through latin1
                    # in the same way the real parser does.
                    headers = HTTPHeaders.parse(
                        native_str(data.decode('latin1')))
                    expected = [('Cookie', 'foo=' +
                                 native_str(encoded.decode('latin1')) + 'bar')]
                    self.assertEqual(
                        expected, list(headers.get_all()))
                except Exception:
                    gen_log.warning("failed while trying %r in %s",
                                    newline, encoding)
                    raise

    def test_optional_cr(self):
        # Both CRLF and LF should be accepted as separators. CR should not be
        # part of the data when followed by LF, but it is a normal char
        # otherwise (or should bare CR be an error?)
        headers = HTTPHeaders.parse(
            'CRLF: crlf\r\nLF: lf\nCR: cr\rMore: more\r\n')
        self.assertEqual(sorted(headers.get_all()),
                         [('Cr', 'cr\rMore: more'),
                          ('Crlf', 'crlf'),
                          ('Lf', 'lf'),
                          ])

    def test_copy(self):
        all_pairs = [('A', '1'), ('A', '2'), ('B', 'c')]
        h1 = HTTPHeaders()
        for k, v in all_pairs:
            h1.add(k, v)
        h2 = h1.copy()
        h3 = copy.copy(h1)
        h4 = copy.deepcopy(h1)
        for headers in [h1, h2, h3, h4]:
            # All the copies are identical, no matter how they were
            # constructed.
            self.assertEqual(list(sorted(headers.get_all())), all_pairs)
        for headers in [h2, h3, h4]:
            # Neither the dict or its member lists are reused.
            self.assertIsNot(headers, h1)
            self.assertIsNot(headers.get_list('A'), h1.get_list('A'))

    def test_pickle_roundtrip(self):
        headers = HTTPHeaders()
        headers.add('Set-Cookie', 'a=b')
        headers.add('Set-Cookie', 'c=d')
        headers.add('Content-Type', 'text/html')
        pickled = pickle.dumps(headers)
        unpickled = pickle.loads(pickled)
        self.assertEqual(sorted(headers.get_all()), sorted(unpickled.get_all()))
        self.assertEqual(sorted(headers.items()), sorted(unpickled.items()))

    def test_setdefault(self):
        headers = HTTPHeaders()
        headers['foo'] = 'bar'
        # If a value is present, setdefault returns it without changes.
        self.assertEqual(headers.setdefault('foo', 'baz'), 'bar')
        self.assertEqual(headers['foo'], 'bar')
        # If a value is not present, setdefault sets it for future use.
        self.assertEqual(headers.setdefault('quux', 'xyzzy'), 'xyzzy')
        self.assertEqual(headers['quux'], 'xyzzy')
        self.assertEqual(sorted(headers.get_all()), [('Foo', 'bar'), ('Quux', 'xyzzy')])

    def test_string(self):
        headers = HTTPHeaders()
        headers.add("Foo", "1")
        headers.add("Foo", "2")
        headers.add("Foo", "3")
        headers2 = HTTPHeaders.parse(str(headers))
        self.assertEquals(headers, headers2)


class FormatTimestampTest(unittest.TestCase):
    # Make sure that all the input types are supported.
    TIMESTAMP = 1359312200.503611
    EXPECTED = 'Sun, 27 Jan 2013 18:43:20 GMT'

    def check(self, value):
        self.assertEqual(format_timestamp(value), self.EXPECTED)

    def test_unix_time_float(self):
        self.check(self.TIMESTAMP)

    def test_unix_time_int(self):
        self.check(int(self.TIMESTAMP))

    def test_struct_time(self):
        self.check(time.gmtime(self.TIMESTAMP))

    def test_time_tuple(self):
        tup = tuple(time.gmtime(self.TIMESTAMP))
        self.assertEqual(9, len(tup))
        self.check(tup)

    def test_datetime(self):
        self.check(datetime.datetime.utcfromtimestamp(self.TIMESTAMP))


# HTTPServerRequest is mainly tested incidentally to the server itself,
# but this tests the parts of the class that can be tested in isolation.
class HTTPServerRequestTest(unittest.TestCase):
    def test_default_constructor(self):
        # All parameters are formally optional, but uri is required
        # (and has been for some time).  This test ensures that no
        # more required parameters slip in.
        HTTPServerRequest(uri='/')

    def test_body_is_a_byte_string(self):
        requets = HTTPServerRequest(uri='/')
        self.assertIsInstance(requets.body, bytes)


class ParseRequestStartLineTest(unittest.TestCase):
    METHOD = "GET"
    PATH = "/foo"
    VERSION = "HTTP/1.1"

    def test_parse_request_start_line(self):
        start_line = " ".join([self.METHOD, self.PATH, self.VERSION])
        parsed_start_line = parse_request_start_line(start_line)
        self.assertEqual(parsed_start_line.method, self.METHOD)
        self.assertEqual(parsed_start_line.path, self.PATH)
        self.assertEqual(parsed_start_line.version, self.VERSION)


class ParseCookieTest(unittest.TestCase):
    # These tests copied from Django:
    # https://github.com/django/django/pull/6277/commits/da810901ada1cae9fc1f018f879f11a7fb467b28
    def test_python_cookies(self):
        """
        Test cases copied from Python's Lib/test/test_http_cookies.py
        """
        self.assertEqual(parse_cookie('chips=ahoy; vienna=finger'), {'chips': 'ahoy', 'vienna': 'finger'})
        # Here parse_cookie() differs from Python's cookie parsing in that it
        # treats all semicolons as delimiters, even within quotes.
        self.assertEqual(
            parse_cookie('keebler="E=mc2; L=\\"Loves\\"; fudge=\\012;"'),
            {'keebler': '"E=mc2', 'L': '\\"Loves\\"', 'fudge': '\\012', '': '"'}
        )
        # Illegal cookies that have an '=' char in an unquoted value.
        self.assertEqual(parse_cookie('keebler=E=mc2'), {'keebler': 'E=mc2'})
        # Cookies with ':' character in their name.
        self.assertEqual(parse_cookie('key:term=value:term'), {'key:term': 'value:term'})
        # Cookies with '[' and ']'.
        self.assertEqual(parse_cookie('a=b; c=[; d=r; f=h'), {'a': 'b', 'c': '[', 'd': 'r', 'f': 'h'})

    def test_cookie_edgecases(self):
        # Cookies that RFC6265 allows.
        self.assertEqual(parse_cookie('a=b; Domain=example.com'), {'a': 'b', 'Domain': 'example.com'})
        # parse_cookie() has historically kept only the last cookie with the
        # same name.
        self.assertEqual(parse_cookie('a=b; h=i; a=c'), {'a': 'c', 'h': 'i'})

    def test_invalid_cookies(self):
        """
        Cookie strings that go against RFC6265 but browsers will send if set
        via document.cookie.
        """
        # Chunks without an equals sign appear as unnamed values per
        # https://bugzilla.mozilla.org/show_bug.cgi?id=169091
        self.assertIn('django_language', parse_cookie('abc=def; unnamed; django_language=en').keys())
        # Even a double quote may be an unamed value.
        self.assertEqual(parse_cookie('a=b; "; c=d'), {'a': 'b', '': '"', 'c': 'd'})
        # Spaces in names and values, and an equals sign in values.
        self.assertEqual(parse_cookie('a b c=d e = f; gh=i'), {'a b c': 'd e = f', 'gh': 'i'})
        # More characters the spec forbids.
        self.assertEqual(parse_cookie('a   b,c<>@:/[]?{}=d  "  =e,f g'), {'a   b,c<>@:/[]?{}': 'd  "  =e,f g'})
        # Unicode characters. The spec only allows ASCII.
        self.assertEqual(parse_cookie('saint=André Bessette'), {'saint': native_str('André Bessette')})
        # Browsers don't send extra whitespace or semicolons in Cookie headers,
        # but parse_cookie() should parse whitespace the same way
        # document.cookie parses whitespace.
        self.assertEqual(parse_cookie('  =  b  ;  ;  =  ;   c  =  ;  '), {'': 'b', 'c': ''})



"""
Manage ruby installations with rbenv. rbenv is supported on Linux and macOS.
rbenv doesn't work on Windows (and isn't really necessary on Windows as there is
no system Ruby on Windows). On Windows, the RubyInstaller and/or Pik are both
good alternatives to work with multiple versions of Ruby on the same box.

http://misheska.com/blog/2013/06/15/using-rbenv-to-manage-multiple-versions-of-ruby/

.. versionadded:: 0.16.0
"""

import logging
import os
import re

import salt.utils.args
import salt.utils.data
import salt.utils.path
import salt.utils.platform
from salt.exceptions import SaltInvocationError

# Set up logger
log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}

__opts__ = {
    "rbenv.root": None,
    "rbenv.build_env": None,
}


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return (
            False,
            "The rbenv execution module failed to load: only available on non-Windows"
            " systems.",
        )
    return True


def _shlex_split(s):
    # from python:salt.utils.args.shlex_split: passing None for s will read
    # the string to split from standard input.
    if s is None:
        ret = salt.utils.args.shlex_split("")
    else:
        ret = salt.utils.args.shlex_split(s)

    return ret


def _parse_env(env):
    if not env:
        env = {}
    if isinstance(env, list):
        env = salt.utils.data.repack_dictlist(env)
    if not isinstance(env, dict):
        env = {}

    for bad_env_key in (x for x, y in env.items() if y is None):
        log.error(
            "Environment variable '%s' passed without a value. "
            "Setting value to an empty string",
            bad_env_key,
        )
        env[bad_env_key] = ""

    return env


def _rbenv_bin(runas=None):
    path = _rbenv_path(runas)
    return f"{path}/bin/rbenv"


def _rbenv_path(runas=None):
    path = None
    if runas in (None, "root"):
        path = __salt__["config.option"]("rbenv.root") or "/usr/local/rbenv"
    else:
        path = __salt__["config.option"]("rbenv.root") or f"~{runas}/.rbenv"

    return os.path.expanduser(path)


def _rbenv_exec(command, env=None, runas=None, ret=None):
    if not is_installed(runas):
        return False

    binary = _rbenv_bin(runas)
    path = _rbenv_path(runas)

    environ = _parse_env(env)
    environ["RBENV_ROOT"] = path

    result = __salt__["cmd.run_all"]([binary] + command, runas=runas, env=environ)

    if isinstance(ret, dict):
        ret.update(result)
        return ret

    if result["retcode"] == 0:
        return result["stdout"]
    else:
        return False


def _install_rbenv(path, runas=None):
    if os.path.isdir(path):
        return True

    cmd = ["git", "clone", "https://github.com/rbenv/rbenv.git", path]
    return __salt__["cmd.retcode"](cmd, runas=runas, python_shell=False) == 0


def _install_ruby_build(path, runas=None):
    path = f"{path}/plugins/ruby-build"
    if os.path.isdir(path):
        return True

    cmd = ["git", "clone", "https://github.com/rbenv/ruby-build.git", path]
    return __salt__["cmd.retcode"](cmd, runas=runas, python_shell=False) == 0


def _update_rbenv(path, runas=None):
    if not os.path.isdir(path):
        return False

    return (
        __salt__["cmd.retcode"](
            ["git", "pull"], runas=runas, cwd=path, python_shell=False
        )
        == 0
    )


def _update_ruby_build(path, runas=None):
    path = f"{path}/plugins/ruby-build"
    if not os.path.isdir(path):
        return False

    return (
        __salt__["cmd.retcode"](
            ["git", "pull"], runas=runas, cwd=path, python_shell=False
        )
        == 0
    )


def install(runas=None, path=None):
    """
    Install rbenv systemwide

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.install
    """
    path = path or _rbenv_path(runas)
    path = os.path.expanduser(path)
    return _install_rbenv(path, runas) and _install_ruby_build(path, runas)


def update(runas=None, path=None):
    """
    Updates the current versions of rbenv and ruby-build

    runas
        The user under which to run rbenv. If not specified, then rbenv will be
        run as the user under which Salt is running.

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.update
    """
    path = path or _rbenv_path(runas)
    path = os.path.expanduser(path)

    return _update_rbenv(path, runas) and _update_ruby_build(path, runas)


def is_installed(runas=None):
    """
    Check if rbenv is installed

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.is_installed
    """
    return __salt__["cmd.has_exec"](_rbenv_bin(runas))


def install_ruby(ruby, runas=None):
    """
    Install a ruby implementation.

    ruby
        The version of Ruby to install, should match one of the
        versions listed by :py:func:`rbenv.list <salt.modules.rbenv.list>`

    runas
        The user under which to run rbenv. If not specified, then rbenv will be
        run as the user under which Salt is running.

    Additional environment variables can be configured in pillar /
    grains / master:

    .. code-block:: yaml

        rbenv:
          build_env: 'CONFIGURE_OPTS="--no-tcmalloc" CFLAGS="-fno-tree-dce"'

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.install_ruby 2.0.0-p0
    """
    ruby = re.sub(r"^ruby-", "", ruby)

    env = None
    env_list = []

    if __grains__["os"] in ("FreeBSD", "NetBSD", "OpenBSD"):
        env_list.append("MAKE=gmake")

    if __salt__["config.get"]("rbenv:build_env"):
        env_list.append(__salt__["config.get"]("rbenv:build_env"))
    elif __salt__["config.option"]("rbenv.build_env"):
        env_list.append(__salt__["config.option"]("rbenv.build_env"))

    if env_list:
        env = " ".join(env_list)

    ret = {}
    ret = _rbenv_exec(["install", ruby], env=env, runas=runas, ret=ret)
    if ret is not False and ret["retcode"] == 0:
        rehash(runas=runas)
        return ret["stderr"]
    else:
        # Cleanup the failed installation so it doesn't list as installed
        uninstall_ruby(ruby, runas=runas)
        return False


def uninstall_ruby(ruby, runas=None):
    """
    Uninstall a ruby implementation.

    ruby
        The version of ruby to uninstall. Should match one of the versions
        listed by :py:func:`rbenv.versions <salt.modules.rbenv.versions>`.

    runas
        The user under which to run rbenv. If not specified, then rbenv will be
        run as the user under which Salt is running.

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.uninstall_ruby 2.0.0-p0
    """
    ruby = re.sub(r"^ruby-", "", ruby)
    _rbenv_exec(["uninstall", "--force", ruby], runas=runas)
    return True


def versions(runas=None):
    """
    List the installed versions of ruby

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.versions
    """
    ret = _rbenv_exec(["versions", "--bare"], runas=runas)
    return [] if ret is False else ret.splitlines()


def default(ruby=None, runas=None):
    """
    Returns or sets the currently defined default ruby

    ruby
        The version to set as the default. Should match one of the versions
        listed by :py:func:`rbenv.versions <salt.modules.rbenv.versions>`.
        Leave blank to return the current default.

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.default
        salt '*' rbenv.default 2.0.0-p0
    """
    if ruby:
        _rbenv_exec(["global", ruby], runas=runas)
        return True
    else:
        ret = _rbenv_exec(["global"], runas=runas)
        return "" if ret is False else ret.strip()


def list_(runas=None):
    """
    List the installable versions of ruby

    runas
        The user under which to run rbenv. If not specified, then rbenv will be
        run as the user under which Salt is running.

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.list
    """
    ret = []
    output = _rbenv_exec(["install", "--list"], runas=runas)
    if output:
        for line in output.splitlines():
            if line == "Available versions:":
                continue
            ret.append(line.strip())
    return ret


def rehash(runas=None):
    """
    Run ``rbenv rehash`` to update the installed shims

    runas
        The user under which to run rbenv. If not specified, then rbenv will be
        run as the user under which Salt is running.

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.rehash
    """
    _rbenv_exec(["rehash"], runas=runas)
    return True


def do(cmdline, runas=None, env=None):
    """
    Execute a ruby command with rbenv's shims from the user or the system

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.do 'gem list bundler'
        salt '*' rbenv.do 'gem list bundler' deploy
    """
    if not cmdline:
        # This is a positional argument so this should never happen, but this
        # will handle cases where someone explicitly passes a false value for
        # cmdline.
        raise SaltInvocationError("Command must be specified")

    path = _rbenv_path(runas)
    if not env:
        env = {}

    # NOTE: Env vars (and their values) need to be str type on both Python 2
    # and 3. The code below first normalizes all path components to unicode to
    # stitch them together, and then converts the result back to a str type.
    env["PATH"] = salt.utils.stringutils.to_str(
        os.pathsep.join(
            (
                salt.utils.path.join(path, "shims"),
                salt.utils.stringutils.to_unicode(os.environ["PATH"]),
            )
        )
    )

    try:
        cmdline = salt.utils.args.shlex_split(cmdline)
    except AttributeError:
        cmdauth = salt.utils.args.shlex_split(str(cmdline))

    result = __salt__["cmd.run_all"](cmdline, runas=runas, env=env, python_shell=False)

    if result["retcode"] == 0:
        rehash(runas=runas)
        return result["stdout"]
    else:
        return False


def do_with_ruby(ruby, cmdline, runas=None):
    """
    Execute a ruby command with rbenv's shims using a specific ruby version

    CLI Example:

    .. code-block:: bash

        salt '*' rbenv.do_with_ruby 2.0.0-p0 'gem list bundler'
        salt '*' rbenv.do_with_ruby 2.0.0-p0 'gem list bundler' runas=deploy
    """
    if not cmdline:
        # This is a positional argument so this should never happen, but this
        # will handle cases where someone explicitly passes a false value for
        # cmdline.
        raise SaltInvocationError("Command must be specified")

    try:
        cmdline = salt.utils.args.shlex_split(cmdline)
    except AttributeError:
        cmdline = salt.utils.args.shlex_split(str(cmdline))

    env = {}
    if ruby:
        env["RBENV_VERSION"] = ruby
        cmd = cmdline
    else:
        cmd = cmdline

    return do(cmd, runas=runas, env=env)



import pytest

import salt.modules.systemd_service as systemd_service
import salt.utils.systemd
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_unless_on_linux,
]


@pytest.fixture
def configure_loader_modules():
    return {systemd_service: {}}


@pytest.mark.parametrize(
    "service_name, expected",
    [
        ("salt-minion", True),
        ("salt-minion.service", True),
        ("other-service", False),
    ],
)
def test_salt_minion_service(service_name, expected):
    """
    Test the _salt_minion_service function to ensure it correctly identifies
    the salt-minion service name.
    """
    assert systemd_service._salt_minion_service(service_name) is expected


@pytest.mark.parametrize(
    "service_name, no_block, expected",
    [
        ("salt-minion", None, True),
        ("other-service", None, False),
        ("salt-minion", False, False),
        ("other-service", False, False),
        ("salt-minion", True, True),
        ("other-service", True, True),
    ],
)
def test_no_block_default(service_name, no_block, expected):
    """
    Test the _no_block_default function to ensure it returns the correct
    default value for the no_block argument based on the service name.
    """
    assert systemd_service._no_block_default(service_name, no_block) is expected


@pytest.mark.skipif(not salt.utils.systemd.booted(), reason="Requires systemd")
@pytest.mark.parametrize(
    "operation,expected_command",
    [
        ("restart", "restart"),
        ("stop", "stop"),
    ],
)
def test_operation_no_block_default(operation, expected_command, grains):
    """
    Test restart/stop functions to ensure they use the correct default value for
    no_block when operating on the salt-minion service.
    """

    if grains["osfinger"] == "Amazon Linux-2":
        pytest.skip("Amazon Linux 2 CI containers do not support systemd fully")

    mock_none = MagicMock(return_value=None)
    mock_true = MagicMock(return_value=True)
    mock_run_all_success = MagicMock(
        return_value={"retcode": 0, "stdout": "", "stderr": "", "pid": 12345}
    )

    with patch.dict(
        systemd_service.__salt__,
        {"cmd.run_all": mock_run_all_success, "config.get": mock_true},
    ) as mock_salt, patch.object(
        systemd_service, "_check_for_unit_changes", mock_none
    ), patch.object(
        systemd_service, "_check_unmask", mock_none
    ), patch(
        "salt.utils.path.which", lambda x: "/usr/bin/" + x
    ):
        # Get the function to test based on the operation parameter
        operation_func = getattr(systemd_service, operation)

        # Test salt-minion.service (should include --no-block)
        assert operation_func("salt-minion.service")
        mock_salt["cmd.run_all"].assert_called_with(
            [
                "/usr/bin/systemd-run",
                "--scope",
                "/usr/bin/systemctl",
                "--no-block",
                expected_command,
                "salt-minion.service",
            ],
            python_shell=False,
        )

        # Test other.service (should not include --no-block)
        assert operation_func("other.service")
        mock_salt["cmd.run_all"].assert_called_with(
            [
                "/usr/bin/systemd-run",
                "--scope",
                "/usr/bin/systemctl",
                expected_command,
                "other.service",
            ],
            python_shell=False,
        )



"""
Module for testing that a __virtual__ function returning False will not be
available via the Salt Loader.
"""


def __virtual__():
    return (False, "The test_virtual execution module failed to load.")


def ping():
    return True



"""
    :codeauthor: :email:`Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`


    salt.config.schemas.esxcluster
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ESX Cluster configuration schemas
"""

from salt.utils.schema import (
    AnyOfItem,
    ArrayItem,
    BooleanItem,
    ComplexSchemaItem,
    DefinitionsSchema,
    DictItem,
    IntegerItem,
    Schema,
    StringItem,
)


class OptionValueItem(ComplexSchemaItem):
    """Sechma item of the OptionValue"""

    title = "OptionValue"
    key = StringItem(title="Key", required=True)
    value = AnyOfItem(items=[StringItem(), BooleanItem(), IntegerItem()])


class AdmissionControlPolicyItem(ComplexSchemaItem):
    """
    Schema item of the HA admission control policy
    """

    title = "Admission Control Policy"

    cpu_failover_percent = IntegerItem(
        title="CPU Failover Percent", minimum=0, maximum=100
    )
    memory_failover_percent = IntegerItem(
        title="Memory Failover Percent", minimum=0, maximum=100
    )


class DefaultVmSettingsItem(ComplexSchemaItem):
    """
    Schema item of the HA default vm settings
    """

    title = "Default VM Settings"

    isolation_response = StringItem(
        title="Isolation Response",
        enum=["clusterIsolationResponse", "none", "powerOff", "shutdown"],
    )
    restart_priority = StringItem(
        title="Restart Priority",
        enum=["clusterRestartPriority", "disabled", "high", "low", "medium"],
    )


class HAConfigItem(ComplexSchemaItem):
    """
    Schema item of ESX cluster high availability
    """

    title = "HA Configuration"
    description = "ESX cluster HA configuration json schema item"

    enabled = BooleanItem(
        title="Enabled", description="Specifies if HA should be enabled"
    )
    admission_control_enabled = BooleanItem(title="Admission Control Enabled")
    admission_control_policy = AdmissionControlPolicyItem()
    default_vm_settings = DefaultVmSettingsItem()
    hb_ds_candidate_policy = StringItem(
        title="Heartbeat Datastore Candidate Policy",
        enum=["allFeasibleDs", "allFeasibleDsWithUserPreference", "userSelectedDs"],
    )
    host_monitoring = StringItem(
        title="Host Monitoring", choices=["enabled", "disabled"]
    )
    options = ArrayItem(min_items=1, items=OptionValueItem())
    vm_monitoring = StringItem(
        title="Vm Monitoring",
        choices=["vmMonitoringDisabled", "vmAndAppMonitoring", "vmMonitoringOnly"],
    )


class vSANClusterConfigItem(ComplexSchemaItem):
    """
    Schema item of the ESX cluster vSAN configuration
    """

    title = "vSAN Configuration"
    description = "ESX cluster vSAN configurationi item"

    enabled = BooleanItem(
        title="Enabled", description="Specifies if vSAN should be enabled"
    )
    auto_claim_storage = BooleanItem(
        title="Auto Claim Storage",
        description=(
            "Specifies whether the storage of member ESXi hosts should "
            "be automatically claimed for vSAN"
        ),
    )
    dedup_enabled = BooleanItem(
        title="Enabled", description="Specifies dedup should be enabled"
    )
    compression_enabled = BooleanItem(
        title="Enabled", description="Specifies if compression should be enabled"
    )


class DRSConfigItem(ComplexSchemaItem):
    """
    Schema item of the ESX cluster DRS configuration
    """

    title = "DRS Configuration"
    description = "ESX cluster DRS configuration item"

    enabled = BooleanItem(
        title="Enabled", description="Specifies if DRS should be enabled"
    )
    vmotion_rate = IntegerItem(
        title="vMotion rate",
        description=(
            "Aggressiveness to do automatic vMotions: "
            "1 (least aggressive) - 5 (most aggressive)"
        ),
        minimum=1,
        maximum=5,
    )
    default_vm_behavior = StringItem(
        title="Default VM DRS Behavior",
        description="Specifies the default VM DRS behavior",
        enum=["fullyAutomated", "partiallyAutomated", "manual"],
    )


class ESXClusterConfigSchema(DefinitionsSchema):
    """
    Schema of the ESX cluster config
    """

    title = "ESX Cluster Configuration Schema"
    description = "ESX cluster configuration schema"

    ha = HAConfigItem()
    vsan = vSANClusterConfigItem()
    drs = DRSConfigItem()
    vm_swap_placement = StringItem(title="VM Swap Placement")


class ESXClusterEntitySchema(Schema):
    """Schema of the ESX cluster entity"""

    title = "ESX Cluster Entity Schema"
    description = "ESX cluster entity schema"

    type = StringItem(
        title="Type",
        description="Specifies the entity type",
        required=True,
        enum=["cluster"],
    )

    datacenter = StringItem(
        title="Datacenter",
        description="Specifies the cluster datacenter",
        required=True,
        pattern=r"\w+",
    )

    cluster = StringItem(
        title="Cluster",
        description="Specifies the cluster name",
        required=True,
        pattern=r"\w+",
    )


class LicenseSchema(Schema):
    """
    Schema item of the ESX cluster vSAN configuration
    """

    title = "Licenses schema"
    description = "License configuration schema"

    licenses = DictItem(
        title="Licenses",
        description="Dictionary containing the license name to key mapping",
        required=True,
        additional_properties=StringItem(
            title="License Key",
            description="Specifies the license key",
            pattern=r"^(\w{5}-\w{5}-\w{5}-\w{5}-\w{5})$",
        ),
    )


class EsxclusterProxySchema(Schema):
    """
    Schema of the esxcluster proxy input
    """

    title = "Esxcluster Proxy Schema"
    description = "Esxcluster proxy schema"
    additional_properties = False
    proxytype = StringItem(required=True, enum=["esxcluster"])
    vcenter = StringItem(required=True, pattern=r"[^\s]+")
    datacenter = StringItem(required=True)
    cluster = StringItem(required=True)
    mechanism = StringItem(required=True, enum=["userpass", "sspi"])
    username = StringItem()
    passwords = ArrayItem(min_items=1, items=StringItem(), unique_items=True)
    # TODO Should be changed when anyOf is supported for schemas
    domain = StringItem()
    principal = StringItem()
    protocol = StringItem()
    port = IntegerItem(minimum=1)



import logging
import os
import shutil
import time

import pytest

from tests.conftest import FIPS_TESTRUN

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.skip_on_freebsd(reason="Processes are not properly killed on FreeBSD"),
]

log = logging.getLogger(__name__)


def test_pki(salt_mm_failover_master_1, salt_mm_failover_master_2, caplog):
    """
    Verify https://docs.saltproject.io/en/latest/topics/tutorials/multimaster_pki.html
    """
    # At first we spin up a simple minion in order to capture its logging output.
    config_defaults = {
        "transport": salt_mm_failover_master_1.config["transport"],
    }

    mm_master_1_port = salt_mm_failover_master_1.config["ret_port"]
    mm_master_1_addr = salt_mm_failover_master_1.config["interface"]
    mm_master_2_port = salt_mm_failover_master_2.config["ret_port"]
    mm_master_2_addr = salt_mm_failover_master_2.config["interface"]
    config_overrides = {
        "master": [
            f"{mm_master_1_addr}:{mm_master_1_port}",
            f"{mm_master_2_addr}:{mm_master_2_port}",
        ],
        "publish_port": salt_mm_failover_master_1.config["publish_port"],
        "master_type": "failover",
        "master_alive_interval": 5,
        "master_tries": -1,
        "verify_master_pubkey_sign": True,
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_mm_failover_master_1.salt_minion_daemon(
        "mm-failover-pki-minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    # Need to grab the public signing key from the master, either will do
    shutil.copyfile(
        os.path.join(salt_mm_failover_master_1.config["pki_dir"], "master_sign.pub"),
        os.path.join(factory.config["pki_dir"], "master_sign.pub"),
    )
    with caplog.at_level(logging.DEBUG):
        with factory.started(start_timeout=120):
            pass

    assert (
        "Successfully verified signature of master public key with verification public key master_sign.pub"
        in caplog.text
    )


def test_return_to_assigned_master(
    mm_failover_master_1_salt_cli,
    mm_failover_master_2_salt_cli,
    salt_mm_failover_minion_1,
    salt_mm_failover_minion_2,
    run_salt_cmds,
):
    """
    Test that values are being returned to only the master the minion is currently connected to.
    """
    returns = run_salt_cmds(
        [mm_failover_master_1_salt_cli, mm_failover_master_2_salt_cli],
        [salt_mm_failover_minion_1, salt_mm_failover_minion_2],
    )

    assert len(returns) == 2
    assert (mm_failover_master_1_salt_cli, salt_mm_failover_minion_1) in returns
    assert (mm_failover_master_2_salt_cli, salt_mm_failover_minion_2) in returns


def test_failover_to_second_master(
    event_listener,
    salt_mm_failover_master_1,
    salt_mm_failover_master_2,
    salt_mm_failover_minion_1,
    salt_mm_failover_minion_2,
    mm_failover_master_1_salt_cli,
    mm_failover_master_2_salt_cli,
    run_salt_cmds,
):
    """
    Test then when the first master is stopped, connected minions failover to the second master.
    """
    event_patterns = [
        (
            salt_mm_failover_master_2.id,
            f"salt/minion/{salt_mm_failover_minion_1.id}/start",
        )
    ]

    start_time = time.time()
    with salt_mm_failover_master_1.stopped():
        assert salt_mm_failover_master_2.is_running()
        # We need to wait for them to realize that the master is not alive
        # At this point, only the first minion will need to change masters
        events = event_listener.wait_for_events(
            event_patterns,
            timeout=salt_mm_failover_minion_1.config["master_alive_interval"] * 4,
            after_time=start_time,
        )

        assert salt_mm_failover_minion_1.is_running()
        assert not events.missed

        returns = run_salt_cmds(
            [mm_failover_master_1_salt_cli, mm_failover_master_2_salt_cli],
            [salt_mm_failover_minion_1, salt_mm_failover_minion_2],
        )

        assert len(returns) == 2
        assert (mm_failover_master_2_salt_cli, salt_mm_failover_minion_1) in returns
        assert (mm_failover_master_2_salt_cli, salt_mm_failover_minion_2) in returns


def test_minion_reconnection(
    salt_mm_failover_minion_1,
    salt_mm_failover_minion_2,
    mm_failover_master_1_salt_cli,
    mm_failover_master_2_salt_cli,
    run_salt_cmds,
):
    """
    Test that minions reconnect to a live master.

    To work well with salt factories, the minions will reconnect to the master they were connected to in conftest.py.
    """
    with salt_mm_failover_minion_1.stopped(), salt_mm_failover_minion_2.stopped():
        log.debug("Minions have stopped. They will restart next.")

    returns = run_salt_cmds(
        [mm_failover_master_1_salt_cli, mm_failover_master_2_salt_cli],
        [salt_mm_failover_minion_1, salt_mm_failover_minion_2],
    )

    assert len(returns) == 2
    assert (mm_failover_master_1_salt_cli, salt_mm_failover_minion_1) in returns
    assert (mm_failover_master_2_salt_cli, salt_mm_failover_minion_2) in returns


@pytest.mark.skip_on_windows
def test_minions_alive_with_no_master(
    grains,
    event_listener,
    salt_mm_failover_master_1,
    salt_mm_failover_master_2,
    salt_mm_failover_minion_1,
    salt_mm_failover_minion_2,
):
    """
    Make sure the minions stay alive after all masters have stopped.
    """
    start_time = time.time()
    with salt_mm_failover_master_1.stopped():
        with salt_mm_failover_master_2.stopped():
            # Make sure they had at least one chance to re-auth
            events = event_listener.wait_for_events(
                [
                    (salt_mm_failover_minion_1.id, "__master_disconnected"),
                    (salt_mm_failover_minion_2.id, "__master_disconnected"),
                ],
                timeout=salt_mm_failover_minion_1.config["master_alive_interval"] * 4,
                after_time=start_time,
            )
            assert not events.missed
            assert salt_mm_failover_minion_1.is_running()
            assert salt_mm_failover_minion_2.is_running()

            start_time = time.time()

    event_patterns = [
        (
            salt_mm_failover_master_1.id,
            f"salt/minion/{salt_mm_failover_minion_1.id}/start",
        ),
        (
            salt_mm_failover_master_1.id,
            f"salt/minion/{salt_mm_failover_minion_2.id}/start",
        ),
        (
            salt_mm_failover_master_2.id,
            f"salt/minion/{salt_mm_failover_minion_1.id}/start",
        ),
        (
            salt_mm_failover_master_2.id,
            f"salt/minion/{salt_mm_failover_minion_2.id}/start",
        ),
    ]
    events = event_listener.wait_for_events(
        event_patterns,
        timeout=salt_mm_failover_minion_1.config["master_alive_interval"] * 8,
        after_time=start_time,
    )

    assert len(events.matches) >= 2

    expected_tags = {
        f"salt/minion/{salt_mm_failover_minion_1.id}/start",
        f"salt/minion/{salt_mm_failover_minion_2.id}/start",
    }
    assert {event.tag for event in events} == expected_tags



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.win_network as win_network
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_network: {}}


def test_managed_missing_parameters():
    """
    Test to ensure that the named interface is configured properly.
    """
    ret = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": (
            "dns_proto must be one of the following: static, dhcp\n"
            "ip_proto must be one of the following: static, dhcp"
        ),
    }
    assert win_network.managed("salt") == ret


def test_managed_static_enabled_false():
    ret = {
        "name": "salt",
        "changes": {},
        "result": True,
        "comment": "Interface 'salt' is up to date (already disabled)",
    }
    mock_false = MagicMock(return_value=False)
    with patch.dict(win_network.__salt__, {"ip.is_enabled": mock_false}):
        assert (
            win_network.managed(
                "salt", dns_proto="static", ip_proto="static", enabled=False
            )
            == ret
        )


def test_managed_test_true():
    ret = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": "Failed to enable interface 'salt' to make changes",
    }
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        win_network.__salt__, {"ip.is_enabled": mock_false, "ip.enable": mock_false}
    ), patch.dict(win_network.__opts__, {"test": False}):
        assert win_network.managed("salt", dns_proto="static", ip_proto="static") == ret


def test_managed_validate_errors():
    ret = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": (
            "The following SLS configuration errors were "
            "detected:\n"
            "- First Error\n"
            "- Second Error"
        ),
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=["First Error", "Second Error"])
    with patch.dict(win_network.__salt__, {"ip.is_enabled": mock_true}), patch.object(
        win_network, "_validate", mock_validate
    ):
        assert win_network.managed("salt", dns_proto="static", ip_proto="static") == ret


def test_managed_get_current_config_failed():
    ret = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": "Unable to get current configuration for interface 'salt'",
    }
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    mock_validate = MagicMock(return_value=[])
    with patch.dict(
        win_network.__salt__,
        {"ip.is_enabled": mock_true, "ip.get_interface": mock_false},
    ), patch.object(win_network, "_validate", mock_validate):

        assert win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp") == ret


def test_managed_test_true_no_changes():
    ret = {
        "name": "salt",
        "changes": {},
        "result": True,
        "comment": "Interface 'salt' is up to date",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        return_value={
            "DHCP enabled": "yes",
            "DNS servers configured through DHCP": "192.168.0.10",
        }
    )
    with patch.dict(
        win_network.__salt__,
        {"ip.is_enabled": mock_true, "ip.get_interface": mock_get_int},
    ), patch.dict(win_network.__opts__, {"test": True}), patch.object(
        win_network, "_validate", mock_validate
    ):
        assert win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp") == ret


def test_managed_test_true_changes():
    ret = {
        "name": "salt",
        "changes": {},
        "result": None,
        "comment": (
            "The following changes will be made to interface "
            "'salt':\n"
            "- DNS protocol will be changed to: dhcp"
        ),
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        return_value={
            "DHCP enabled": "no",
            "Statically Configured DNS Servers": "192.168.0.10",
        }
    )
    with patch.dict(
        win_network.__salt__,
        {"ip.is_enabled": mock_true, "ip.get_interface": mock_get_int},
    ), patch.dict(win_network.__opts__, {"test": True}), patch.object(
        win_network, "_validate", mock_validate
    ):

        assert win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp") == ret


def test_managed_failed():
    ret = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": "Failed to set desired configuration settings for interface 'salt'",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        return_value={
            "DHCP enabled": "no",
            "Statically Configured DNS Servers": "192.168.0.10",
        }
    )
    with patch.dict(
        win_network.__salt__,
        {
            "ip.is_enabled": mock_true,
            "ip.get_interface": mock_get_int,
            "ip.set_dhcp_dns": mock_true,
            "ip.set_dhcp_ip": mock_true,
        },
    ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
        win_network, "_validate", mock_validate
    ):
        assert win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp") == ret


def test_managed():
    ret = {
        "name": "salt",
        "changes": {
            "DHCP enabled": {"new": "yes", "old": "no"},
            "DNS servers configured through DHCP": {"new": "192.168.0.10", "old": ""},
            "Statically Configured DNS Servers": {"new": "", "old": "192.168.0.10"},
        },
        "result": True,
        "comment": "Successfully updated configuration for interface 'salt'",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        side_effect=[
            {
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            },
            {
                "DHCP enabled": "yes",
                "DNS servers configured through DHCP": "192.168.0.10",
            },
        ]
    )
    with patch.dict(
        win_network.__salt__,
        {
            "ip.is_enabled": mock_true,
            "ip.get_interface": mock_get_int,
            "ip.set_dhcp_dns": mock_true,
            "ip.set_dhcp_ip": mock_true,
        },
    ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
        win_network, "_validate", mock_validate
    ):
        assert win_network.managed("salt", dns_proto="dhcp", ip_proto="dhcp") == ret


def test_managed_static_dns_clear():
    expected = {
        "name": "salt",
        "changes": {
            "Statically Configured DNS Servers": {"new": "None", "old": "192.168.0.10"}
        },
        "result": True,
        "comment": "Successfully updated configuration for interface 'salt'",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        side_effect=[
            {
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            },
            {"DHCP enabled": "no", "Statically Configured DNS Servers": "None"},
        ]
    )
    with patch.dict(
        win_network.__salt__,
        {
            "ip.is_enabled": mock_true,
            "ip.get_interface": mock_get_int,
            "ip.set_static_dns": mock_true,
        },
    ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
        win_network, "_validate", mock_validate
    ):
        ret = win_network.managed(
            "salt", dns_proto="static", dns_servers=[], ip_proto="dhcp"
        )
        assert ret == expected


def test_managed_static_dns():
    expected = {
        "name": "salt",
        "changes": {
            "Statically Configured DNS Servers": {"new": "192.168.0.10", "old": "None"}
        },
        "result": True,
        "comment": "Successfully updated configuration for interface 'salt'",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        side_effect=[
            {"DHCP enabled": "no", "Statically Configured DNS Servers": "None"},
            {
                "DHCP enabled": "no",
                "Statically Configured DNS Servers": "192.168.0.10",
            },
        ]
    )
    with patch.dict(
        win_network.__salt__,
        {
            "ip.is_enabled": mock_true,
            "ip.get_interface": mock_get_int,
            "ip.set_static_dns": mock_true,
        },
    ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
        win_network, "_validate", mock_validate
    ):
        ret = win_network.managed(
            "salt",
            dns_proto="static",
            dns_servers=["192.168.0.10"],
            ip_proto="dhcp",
        )
        assert ret == expected


def test_managed_static_dns_no_action():
    expected = {
        "name": "salt",
        "changes": {},
        "result": True,
        "comment": "Interface 'salt' is up to date",
    }
    mock_true = MagicMock(return_value=True)
    mock_validate = MagicMock(return_value=[])
    mock_get_int = MagicMock(
        return_value={
            "DHCP enabled": "no",
            "Statically Configured DNS Servers": "192.168.0.10",
        }
    )
    with patch.dict(
        win_network.__salt__,
        {
            "ip.is_enabled": mock_true,
            "ip.get_interface": mock_get_int,
            "ip.set_static_dns": mock_true,
        },
    ), patch.dict(win_network.__opts__, {"test": False}), patch.object(
        win_network, "_validate", mock_validate
    ):
        # Don't pass dns_servers
        ret = win_network.managed("salt", dns_proto="static", ip_proto="dhcp")
        assert ret == expected
        # Pass dns_servers=None
        ret = win_network.managed(
            "salt", dns_proto="static", dns_servers=None, ip_proto="dhcp"
        )
        assert ret == expected



import pytest

import salt.states.ethtool as ethtool
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        ethtool: {
            "__opts__": {"test": False},
            "__salt__": {},
        }
    }


def test_ethtool_pause():
    expected = {
        "changes": {},
        "comment": "Network device eth0 pause parameters are up to date.",
        "name": "eth0",
        "result": True,
    }
    show_ret = {
        "Autonegotiate": True,
        "RX": True,
        "RX negotiated": False,
        "TX": True,
        "TX negotiated": False,
    }
    mock_show = MagicMock(return_value=show_ret)
    mock_set = MagicMock(return_value=True)
    with patch.dict(
        ethtool.__salt__,
        {"ethtool.set_pause": mock_set, "ethtool.show_pause": mock_show},
    ):
        # clean
        ret = ethtool.pause("eth0", autoneg=True, rx=True, tx=True)
        assert ret == expected

        # changes
        expected["changes"] = {"ethtool_pause": "autoneg: off\nrx: off\ntx: off"}
        expected["comment"] = "Device eth0 pause parameters updated."
        ret = ethtool.pause("eth0", autoneg=False, rx=False, tx=False)
        assert ret == expected
        mock_set.assert_called_once_with("eth0", autoneg=False, rx=False, tx=False)

        # changes, test mode
        mock_set.reset_mock()
        with patch.dict(ethtool.__opts__, {"test": True}):
            expected["result"] = None
            expected["changes"] = {}
            expected["comment"] = (
                "Device eth0 pause parameters are set to be updated:\nautoneg: off\nrx: off\ntx: off"
            )
            ret = ethtool.pause("eth0", autoneg=False, rx=False, tx=False)
            assert ret == expected
            mock_set.assert_not_called()

    # exceptions
    with patch.dict(
        ethtool.__salt__,
        {
            "ethtool.set_pause": MagicMock(side_effect=CommandExecutionError("blargh")),
            "ethtool.show_pause": MagicMock(
                side_effect=[CommandExecutionError, show_ret]
            ),
        },
    ):
        expected["comment"] = "Device eth0 pause parameters are not supported"
        expected["result"] = False
        ret = ethtool.pause("eth0", autoneg=False, rx=False, tx=False)
        assert ret == expected
        ret = ethtool.pause("eth0", autoneg=False, rx=False, tx=False)
        expected["comment"] = "blargh"
        assert ret == expected



"""
Connection module for Amazon Elasticsearch Service

.. versionadded:: 2016.11.0

:configuration: This module accepts explicit AWS credentials but can also
    utilize IAM roles assigned to the instance trough Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        lambda.keyid: GKTADJGHEIQSXMKKRBJ08H
        lambda.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        lambda.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

    Create and delete methods return:

    .. code-block:: yaml

        created: true

    or

    .. code-block:: yaml

        created: false
        error:
          message: error message

    Request methods (e.g., `describe_function`) return:

    .. code-block:: yaml

        domain:
          - {...}
          - {...}

    or

    .. code-block:: yaml

        error:
          message: error message

:depends: boto3

"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging

import salt.utils.compat
import salt.utils.json
import salt.utils.versions
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


# pylint: disable=import-error
try:
    # pylint: disable=unused-import
    import boto
    import boto3

    # pylint: enable=unused-import
    from botocore.exceptions import ClientError

    logging.getLogger("boto").setLevel(logging.CRITICAL)
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=import-error


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    # the boto_lambda execution module relies on the connect_to_region() method
    # which was added in boto 2.8.0
    # https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
    return salt.utils.versions.check_boto_reqs(boto_ver="2.8.0", boto3_ver="1.4.0")


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto3.assign_funcs"](__name__, "es")


def exists(DomainName, region=None, key=None, keyid=None, profile=None):
    """
    Given a domain name, check to see if the given domain exists.

    Returns True if the given domain exists and returns False if the given
    function does not exist.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.exists mydomain

    """

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        domain = conn.describe_elasticsearch_domain(DomainName=DomainName)
        return {"exists": True}
    except ClientError as e:
        if e.response.get("Error", {}).get("Code") == "ResourceNotFoundException":
            return {"exists": False}
        return {"error": __utils__["boto3.get_error"](e)}


def status(DomainName, region=None, key=None, keyid=None, profile=None):
    """
    Given a domain name describe its status.

    Returns a dictionary of interesting properties.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.status mydomain

    """

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        domain = conn.describe_elasticsearch_domain(DomainName=DomainName)
        if domain and "DomainStatus" in domain:
            domain = domain.get("DomainStatus", {})
            keys = (
                "Endpoint",
                "Created",
                "Deleted",
                "DomainName",
                "DomainId",
                "EBSOptions",
                "SnapshotOptions",
                "AccessPolicies",
                "Processing",
                "AdvancedOptions",
                "ARN",
                "ElasticsearchVersion",
            )
            return {"domain": {k: domain.get(k) for k in keys if k in domain}}
        else:
            return {"domain": None}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def describe(DomainName, region=None, key=None, keyid=None, profile=None):
    """
    Given a domain name describe its properties.

    Returns a dictionary of interesting properties.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.describe mydomain

    """

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        domain = conn.describe_elasticsearch_domain_config(DomainName=DomainName)
        if domain and "DomainConfig" in domain:
            domain = domain["DomainConfig"]
            keys = (
                "ElasticsearchClusterConfig",
                "EBSOptions",
                "AccessPolicies",
                "SnapshotOptions",
                "AdvancedOptions",
            )
            return {
                "domain": {
                    k: domain.get(k, {}).get("Options") for k in keys if k in domain
                }
            }
        else:
            return {"domain": None}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def create(
    DomainName,
    ElasticsearchClusterConfig=None,
    EBSOptions=None,
    AccessPolicies=None,
    SnapshotOptions=None,
    AdvancedOptions=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    ElasticsearchVersion=None,
):
    """
    Given a valid config, create a domain.

    Returns {created: true} if the domain was created and returns
    {created: False} if the domain was not created.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.create mydomain \\
              {'InstanceType': 't2.micro.elasticsearch', 'InstanceCount': 1, \\
              'DedicatedMasterEnabled': false, 'ZoneAwarenessEnabled': false} \\
              {'EBSEnabled': true, 'VolumeType': 'gp2', 'VolumeSize': 10, \\
              'Iops': 0} \\
              {"Version": "2012-10-17", "Statement": [{"Effect": "Allow", "Principal": {"AWS": "*"}, "Action": "es:*", \\
               "Resource": "arn:aws:es:us-east-1:111111111111:domain/mydomain/*", \\
               "Condition": {"IpAddress": {"aws:SourceIp": ["127.0.0.1"]}}}]} \\
              {"AutomatedSnapshotStartHour": 0} \\
              {"rest.action.multi.allow_explicit_index": "true"}
    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        kwargs = {}
        for k in (
            "ElasticsearchClusterConfig",
            "EBSOptions",
            "AccessPolicies",
            "SnapshotOptions",
            "AdvancedOptions",
            "ElasticsearchVersion",
        ):
            if locals()[k] is not None:
                val = locals()[k]
                if isinstance(val, str):
                    try:
                        val = salt.utils.json.loads(val)
                    except ValueError as e:
                        return {
                            "updated": False,
                            "error": f"Error parsing {k}: {e.message}",
                        }
                kwargs[k] = val
        if "AccessPolicies" in kwargs:
            kwargs["AccessPolicies"] = salt.utils.json.dumps(kwargs["AccessPolicies"])
        if "ElasticsearchVersion" in kwargs:
            kwargs["ElasticsearchVersion"] = str(kwargs["ElasticsearchVersion"])
        domain = conn.create_elasticsearch_domain(DomainName=DomainName, **kwargs)
        if domain and "DomainStatus" in domain:
            return {"created": True}
        else:
            log.warning("Domain was not created")
            return {"created": False}
    except ClientError as e:
        return {"created": False, "error": __utils__["boto3.get_error"](e)}


def delete(DomainName, region=None, key=None, keyid=None, profile=None):
    """
    Given a domain name, delete it.

    Returns {deleted: true} if the domain was deleted and returns
    {deleted: false} if the domain was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.delete mydomain

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_elasticsearch_domain(DomainName=DomainName)
        return {"deleted": True}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def update(
    DomainName,
    ElasticsearchClusterConfig=None,
    EBSOptions=None,
    AccessPolicies=None,
    SnapshotOptions=None,
    AdvancedOptions=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Update the named domain to the configuration.

    Returns {updated: true} if the domain was updated and returns
    {updated: False} if the domain was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.update mydomain \\
              {'InstanceType': 't2.micro.elasticsearch', 'InstanceCount': 1, \\
              'DedicatedMasterEnabled': false, 'ZoneAwarenessEnabled': false} \\
              {'EBSEnabled': true, 'VolumeType': 'gp2', 'VolumeSize': 10, \\
              'Iops': 0} \\
              {"Version": "2012-10-17", "Statement": [{"Effect": "Allow", "Principal": {"AWS": "*"}, "Action": "es:*", \\
               "Resource": "arn:aws:es:us-east-1:111111111111:domain/mydomain/*", \\
               "Condition": {"IpAddress": {"aws:SourceIp": ["127.0.0.1"]}}}]} \\
              {"AutomatedSnapshotStartHour": 0} \\
              {"rest.action.multi.allow_explicit_index": "true"}

    """

    call_args = {}
    for k in (
        "ElasticsearchClusterConfig",
        "EBSOptions",
        "AccessPolicies",
        "SnapshotOptions",
        "AdvancedOptions",
    ):
        if locals()[k] is not None:
            val = locals()[k]
            if isinstance(val, str):
                try:
                    val = salt.utils.json.loads(val)
                except ValueError as e:
                    return {
                        "updated": False,
                        "error": f"Error parsing {k}: {e.message}",
                    }
            call_args[k] = val
    if "AccessPolicies" in call_args:
        call_args["AccessPolicies"] = salt.utils.json.dumps(call_args["AccessPolicies"])
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        domain = conn.update_elasticsearch_domain_config(
            DomainName=DomainName, **call_args
        )
        if not domain or "DomainConfig" not in domain:
            log.warning("Domain was not updated")
            return {"updated": False}
        return {"updated": True}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def add_tags(
    DomainName=None, ARN=None, region=None, key=None, keyid=None, profile=None, **kwargs
):
    """
    Add tags to a domain

    Returns {tagged: true} if the domain was tagged and returns
    {tagged: False} if the domain was not tagged.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elasticsearch_domain.add_tags mydomain tag_a=tag_value tag_b=tag_value

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        tagslist = []
        for k, v in kwargs.items():
            if str(k).startswith("__"):
                continue
            tagslist.append({"Key": str(k), "Value": str(v)})
        if ARN is None:
            if DomainName is None:
                raise SaltInvocationError(
                    "One (but not both) of ARN or domain must be specified."
                )
            domaindata = status(
                DomainName=DomainName,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not domaindata or "domain" not in domaindata:
                log.warning("Domain tags not updated")
                return {"tagged": False}
            ARN = domaindata.get("domain", {}).get("ARN")
        elif DomainName is not None:
            raise SaltInvocationError(
                "One (but not both) of ARN or domain must be specified."
            )
        conn.add_tags(ARN=ARN, TagList=tagslist)
        return {"tagged": True}
    except ClientError as e:
        return {"tagged": False, "error": __utils__["boto3.get_error"](e)}


def remove_tags(
    TagKeys, DomainName=None, ARN=None, region=None, key=None, keyid=None, profile=None
):
    """
    Remove tags from a trail

    Returns {tagged: true} if the trail was tagged and returns
    {tagged: False} if the trail was not tagged.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.remove_tags my_trail tag_a=tag_value tag_b=tag_value

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if ARN is None:
            if DomainName is None:
                raise SaltInvocationError(
                    "One (but not both) of ARN or domain must be specified."
                )
            domaindata = status(
                DomainName=DomainName,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not domaindata or "domain" not in domaindata:
                log.warning("Domain tags not updated")
                return {"tagged": False}
            ARN = domaindata.get("domain", {}).get("ARN")
        elif DomainName is not None:
            raise SaltInvocationError(
                "One (but not both) of ARN or domain must be specified."
            )
        conn.remove_tags(ARN=domaindata.get("domain", {}).get("ARN"), TagKeys=TagKeys)
        return {"tagged": True}
    except ClientError as e:
        return {"tagged": False, "error": __utils__["boto3.get_error"](e)}


def list_tags(
    DomainName=None, ARN=None, region=None, key=None, keyid=None, profile=None
):
    """
    List tags of a trail

    Returns:
        tags:
          - {...}
          - {...}

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudtrail.list_tags my_trail

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if ARN is None:
            if DomainName is None:
                raise SaltInvocationError(
                    "One (but not both) of ARN or domain must be specified."
                )
            domaindata = status(
                DomainName=DomainName,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not domaindata or "domain" not in domaindata:
                log.warning("Domain tags not updated")
                return {"tagged": False}
            ARN = domaindata.get("domain", {}).get("ARN")
        elif DomainName is not None:
            raise SaltInvocationError(
                "One (but not both) of ARN or domain must be specified."
            )
        ret = conn.list_tags(ARN=ARN)
        log.warning(ret)
        tlist = ret.get("TagList", [])
        tagdict = {}
        for tag in tlist:
            tagdict[tag.get("Key")] = tag.get("Value")
        return {"tags": tagdict}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}



"""
Module to provide information about minions and handle killing and restarting
minions
"""

import datetime
import logging
import os
import sys
import time

import salt.key
import salt.utils.data
import salt.utils.path
import salt.utils.systemd
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

# Don't shadow built-ins.
__func_alias__ = {"list_": "list"}


def _is_systemd_system():
    """
    Check if the system uses systemd
    """
    is_linux = __grains__.get("kernel") == "Linux"
    is_booted = salt.utils.systemd.booted(__context__)
    if is_linux and is_booted:
        return True
    return False


def _is_windows_system():
    """
    Check if the system is Windows
    """
    return __grains__.get("kernel") == "Windows"


def _schedule_retry_systemd(retry_delay):
    """
    Schedule a retry for the minion restart on systemd systems
    """
    if not _is_systemd_system():
        return False

    # systemd-run --on-active=180 bash -c "systemctl is-active salt-minion || systemctl start salt-minion"
    cmd = [
        salt.utils.path.which("systemd-run"),
        f"--on-active={retry_delay}",
        "/bin/sh",
        "-c",
        "systemctl is-active salt-minion || systemctl start salt-minion",
    ]

    out = __salt__["cmd.run_all"](cmd, python_shell=False)
    if out["retcode"] != 0:
        raise CommandExecutionError(out["stderr"])
    return out


def _schedule_retry_windows(retry_delay):
    """
    Schedule a retry for the minion restart on Windows systems
    """
    if not _is_windows_system():
        return False

    when = datetime.datetime.now() + datetime.timedelta(seconds=retry_delay)
    cmd = salt.utils.path.which("powershell.exe")
    args = "-ExecutionPolicy Bypass -NoProfile -WindowStyle Hidden -Command \"if ((Get-Service -Name salt-minion).Status -ne 'Running') { Start-Service -Name salt-minion }\""
    out = __salt__["task.create_task"](
        name="retry-minion-restart",
        user_name="System",
        force=True,
        action_type="Execute",
        cmd=cmd,
        arguments=args,
        trigger_type="Once",
        start_date=when.strftime("%Y-%m-%d"),
        start_time=when.strftime("%H:%M:%S"),
    )
    if not out:
        raise CommandExecutionError("Failed to add retry task")
    return out


def _schedule_retry(retry_delay):
    """
    Schedule a retry for the minion restart

    """

    try:
        int(retry_delay)
    except (ValueError, TypeError):
        raise CommandExecutionError(
            "Invalid retry_delay value: {}. Must be a number of seconds.".format(
                retry_delay
            )
        )
    if _is_systemd_system():
        return _schedule_retry_systemd(retry_delay)
    elif _is_windows_system():
        return _schedule_retry_windows(retry_delay)
    return False


def _restart_service(
    service_name, schedule_retry, retry_delay, comment, no_block=False
):
    """
    Helper function to restart a service with optional retry scheduling
    """
    ret = {"service_restart": {}}
    schedule_retry_failed = False

    try:
        if schedule_retry:
            ret["service_restart"]["schedule_retry"] = {}

            schedule_retry_failed = True
            sched = _schedule_retry(retry_delay)
            schedule_retry_failed = False

            if _is_systemd_system():
                ret["service_restart"]["schedule_retry"]["detail"] = sched.get(
                    "stderr", ""
                )
            ret["service_restart"]["schedule_retry"]["delay"] = retry_delay
            comment.append(
                f"Scheduled retry for minion restart in {retry_delay} seconds"
            )

        # Restart the service
        if no_block:
            ret["service_restart"]["result"] = __salt__["service.restart"](
                service_name, no_block=True
            )
        else:
            ret["service_restart"]["result"] = __salt__["service.restart"](service_name)
            if not ret["service_restart"]["result"]:
                raise CommandExecutionError(f"Failed to restart {service_name} service")

        comment.append("Service restart successful")
        ret["retcode"] = 0

    except CommandExecutionError as e:
        comment.append(
            "Adding scheduled retry failed"
            if schedule_retry_failed
            else "Service restart failed"
        )
        ret["service_restart"]["result"] = False
        ret["service_restart"]["stderr"] = str(e)
        ret["retcode"] = salt.defaults.exitcodes.EX_SOFTWARE

    return ret


def list_():
    """
    Return a list of accepted, denied, unaccepted and rejected keys.
    This is the same output as `salt-key -L`

    CLI Example:

    .. code-block:: bash

        salt 'master' minion.list
    """
    pki_dir = __salt__["config.get"]("pki_dir", "")

    # We have to replace the minion/master directories
    pki_dir = pki_dir.replace("minion", "master")

    # The source code below is (nearly) a copy of salt.key.Key.list_keys
    key_dirs = _check_minions_directories(pki_dir)

    ret = {}

    for dir_ in key_dirs:
        ret[os.path.basename(dir_)] = []
        try:
            for fn_ in salt.utils.data.sorted_ignorecase(os.listdir(dir_)):
                if not fn_.startswith("."):
                    if os.path.isfile(os.path.join(dir_, fn_)):
                        ret[os.path.basename(dir_)].append(fn_)
        except OSError:
            # key dir kind is not created yet, just skip
            continue

    return ret


def _check_minions_directories(pki_dir):
    """
    Return the minion keys directory paths.

    This function is a copy of salt.key.Key._check_minions_directories.
    """
    minions_accepted = os.path.join(pki_dir, salt.key.Key.ACC)
    minions_pre = os.path.join(pki_dir, salt.key.Key.PEND)
    minions_rejected = os.path.join(pki_dir, salt.key.Key.REJ)
    minions_denied = os.path.join(pki_dir, salt.key.Key.DEN)

    return minions_accepted, minions_pre, minions_rejected, minions_denied


def kill(timeout=15):
    """
    Kill the salt minion.

    timeout
        int seconds to wait for the minion to die.

    If you have a monitor that restarts ``salt-minion`` when it dies then this is
    a great way to restart after a minion upgrade.

    CLI Example:

    .. code-block:: bash

        salt minion[12] minion.kill

        minion1:
            ----------
            killed:
                7874
            retcode:
                0
        minion2:
            ----------
            killed:
                29071
            retcode:
                0

    The result of the salt command shows the process ID of the minions and the
    results of a kill signal to the minion in as the ``retcode`` value: ``0``
    is success, anything else is a failure.
    """

    ret = {
        "killed": None,
        "retcode": 1,
    }
    comment = []
    pid = __grains__.get("pid")
    if not pid:
        comment.append('Unable to find "pid" in grains')
        ret["retcode"] = salt.defaults.exitcodes.EX_SOFTWARE
    else:
        if "ps.kill_pid" not in __salt__:
            comment.append("Missing command: ps.kill_pid")
            ret["retcode"] = salt.defaults.exitcodes.EX_SOFTWARE
        else:
            # The retcode status comes from the first kill signal
            ret["retcode"] = int(not __salt__["ps.kill_pid"](pid))

            # If the signal was successfully delivered then wait for the
            # process to die - check by sending signals until signal delivery
            # fails.
            if ret["retcode"]:
                comment.append("ps.kill_pid failed")
            else:
                for _ in range(timeout):
                    time.sleep(1)
                    signaled = __salt__["ps.kill_pid"](pid)
                    if not signaled:
                        ret["killed"] = pid
                        break
                else:
                    # The process did not exit before the timeout
                    comment.append("Timed out waiting for minion to exit")
                    ret["retcode"] = salt.defaults.exitcodes.EX_TEMPFAIL

    if comment:
        ret["comment"] = comment
    return ret


def restart(systemd=True, win_service=True, schedule_retry=False, retry_delay=180):
    """
    Restart the salt minion.

    The method to restart the minion will be chosen as follows:

        If ``minion_restart_command`` is set in the minion configuration then
        the command specified will be used to restart the minion.

        If the minion is running as a systemd service then the minion will be
        restarted using the systemd_service module, unless ``systemd`` is
        set to ``False``

        If the minion is running as a Windows service then the minion will be
        restarted using the win_service module, unless ``win_service`` is
        set to ``False``

        If the salt-minion process is running in daemon mode (the ``-d``
        argument is present in ``argv``) then the minion will be killed and
        restarted using the same command line arguments, if possible.

        If the salt-minion process is running in the foreground (the ``-d``
        argument is not present in ``argv``) then the minion will be killed but
        not restarted. This behavior is intended for minion processes that are
        managed by a process supervisor.

    systemd
        If set to ``False`` then systemd will not be used to restart the minion.
        Defaults to ``True``.

    win_service
        If set to ``False`` then the Windows service manager will not be used to
        restart the minion. Defaults to ``True``.

    schedule_retry
        If set to ``True`` then a scheduled job will be added to start the
        minion if it has failed to restart after the retry_delay

    retry_delay
        The amount of time to wait before attempting to start the minion if it
        has failed to restart. Defaults to 180 seconds.

    CLI Examples:

    .. code-block:: bash

        salt minion[12] minion.restart

        minion1:
        ----------
            comment:
                - Using systemctl to restart salt-minion
                - Service restart successful
            killed:
                None
            restart:
                ----------
            retcode:
                0
            service_restart:
                ----------
                result:
                    True
        minion2:
            ----------
            comment:
                - Using windows service manager to restart salt-minion
                - Service restart successful
            killed:
                None
            restart:
                ----------
            retcode:
                0
            service_restart:
                ----------
                result:
                    True

    The result shows that ``minion1`` was restarted using systemd and
    ``minion2`` was restarted using the Windows service manager. The
    ``service_restart`` field indicates the result of the service restart
    operation. The ``killed`` field is ``None`` because the minion was restarted
    using the service manager and not by killing the process. The ``restart``
    field is empty because the minion was restarted using the service manager
    and not by running the command line arguments of the minion process.

    .. code-block:: bash

        salt minion[12] minion.restart

        minion1:
            ----------
            comment:
                - Restart using process argv:
                -     /home/omniture/install/bin/salt-minion
                -     -d
                -     -c
                -     /home/omniture/install/etc/salt
            killed:
                10070
            restart:
                ----------
                stderr:
                stdout:
            retcode:
                0
        minion2:
            ----------
            comment:
                - Using configuration minion_restart_command:
                -     /home/omniture/install/bin/salt-minion
                -     --not-an-option
                -     -d
                -     -c
                -     /home/omniture/install/etc/salt
                - Restart failed
            killed:
                10896
            restart:
                ----------
                stderr:
                    Usage: salt-minion

                    salt-minion: error: no such option: --not-an-option
                stdout:
            retcode:
                64

    The result of the command shows the process ID of ``minion1`` that is
    shutdown (killed) and the results of the restart.  If there is a failure
    in the restart it will be reflected in a non-zero ``retcode`` and possibly
    output in the ``stderr`` and/or ``stdout`` values along with addition
    information in the ``comment`` field as is demonstrated with ``minion2``.
    """

    should_kill = True
    should_restart = True
    comment = []
    ret = {
        "killed": None,
        "restart": {},
        "service_restart": {},
        "retcode": 0,
    }

    restart_cmd = __salt__["config.get"]("minion_restart_command")
    if restart_cmd:
        comment.append("Using configuration minion_restart_command:")
        comment.extend([f"    {arg}" for arg in restart_cmd])
    elif systemd and _is_systemd_system():
        # If we are using systemd then we will restart the minion using
        # service.restart (systemd_service.restart)
        comment.append("Using systemctl to restart salt-minion")
        should_kill = False
        should_restart = False
    elif win_service and _is_windows_system():
        # If we are on Windows then we will restart the minion using
        # service.restart (win_service.restart)
        comment.append("Using windows service manager to restart salt-minion")
        should_kill = False
        should_restart = False
    else:
        if "-d" in sys.argv:
            restart_cmd = sys.argv
            comment.append("Restart using process argv:")
            comment.extend([f"    {arg}" for arg in restart_cmd])
        else:
            should_restart = False
            comment.append(
                "Not running in daemon mode - will not restart process after killing"
            )

    if should_kill:
        ret.update(kill())
        if "comment" in ret and ret["comment"]:
            if isinstance(ret["comment"], str):
                comment.append(ret["comment"])
            else:
                comment.extend(ret["comment"])
        if ret["retcode"]:
            comment.append("Kill failed - not restarting")
            should_restart = False

    if should_restart:
        ret["restart"] = __salt__["cmd.run_all"](restart_cmd, env=os.environ)
        # Do not want to mislead users to think that the returned PID from
        # cmd.run_all() is the PID of the new salt minion - just delete the
        # returned PID.
        if "pid" in ret["restart"]:
            del ret["restart"]["pid"]
        if ret["restart"].get("retcode", None):
            comment.append("Restart failed")
            ret["retcode"] = ret["restart"]["retcode"]
        if "retcode" in ret["restart"]:
            # Just want a single retcode
            del ret["restart"]["retcode"]

    if not restart_cmd:
        if systemd and _is_systemd_system():
            service_result = _restart_service(
                "salt-minion", schedule_retry, retry_delay, comment, no_block=True
            )
            ret.update(service_result)
        elif win_service and _is_windows_system():
            service_result = _restart_service(
                "salt-minion", schedule_retry, retry_delay, comment
            )
            ret.update(service_result)

    if comment:
        ret["comment"] = comment

    return ret



"""
    :codeauthor: `Nitin Madhok <nmadhok@g.clemson.edu>`

    tests.unit.cloud.clouds.vmware_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

from salt import config
from salt.cloud.clouds import vmware
from salt.exceptions import SaltCloudSystemExit
from tests.support.mock import MagicMock, Mock, patch

# Attempt to import pyVim and pyVmomi libs
HAS_LIBS = True
# pylint: disable=import-error,no-name-in-module,unused-import
try:
    from pyVim.connect import Disconnect, SmartConnect
    from pyVmomi import vim, vmodl
except ImportError:
    HAS_LIBS = False
# pylint: enable=import-error,no-name-in-module,unused-import


pytestmark = [
    pytest.mark.skipif(
        not HAS_LIBS, reason="Install pyVmomi to be able to run this test."
    )
]


@pytest.fixture
def vm_name():
    return "test-vm"


@pytest.fixture
def profile():
    return {
        "base-gold": {
            "provider": "vcenter01:vmware",
            "datastore": "Datastore1",
            "resourcepool": "Resources",
            "folder": "vm",
        }
    }


@pytest.fixture
def configure_loader_modules(profile):
    return {
        vmware: {
            "__active_provider_name__": "",
            "__opts__": {
                "providers": {
                    "vcenter01": {
                        "vmware": {
                            "driver": "vmware",
                            "url": "vcenter01.domain.com",
                            "user": "DOMAIN\\user",
                            "password": "verybadpass",
                            "profiles": profile,
                        }
                    }
                }
            },
        }
    }


def test_test_vcenter_connection_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call test_vcenter_connection
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.test_vcenter_connection, call="action")


def test_get_vcenter_version_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call get_vcenter_version
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.get_vcenter_version, call="action")


def test_avail_images_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_images
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.avail_images, call="action")


def test_avail_locations_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_locations
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.avail_locations, call="action")


def test_avail_sizes_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_sizes
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.avail_sizes, call="action")


def test_list_datacenters_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_datacenters
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_datacenters, call="action")


def test_list_clusters_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_clusters
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_clusters, call="action")


def test_list_datastore_clusters_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_datastore_clusters
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_datastore_clusters, call="action")


def test_list_datastores_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_datastores
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_datastores, call="action")


def test_list_hosts_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_hosts
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_hosts, call="action")


def test_list_resourcepools_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_resourcepools
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_resourcepools, call="action")


def test_list_networks_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_networks
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_networks, call="action")


def test_list_nodes_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_nodes
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_nodes, call="action")


def test_list_nodes_min_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_nodes_min
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_nodes_min, call="action")


def test_list_nodes_full_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_nodes_full
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_nodes_full, call="action")


def test_list_nodes_select_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_nodes_full
    with --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_nodes_select, call="action")


def test_list_folders_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_folders
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_folders, call="action")


def test_list_snapshots_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_snapshots
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_snapshots, call="action")


def test_list_hosts_by_cluster_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_hosts_by_cluster
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_hosts_by_cluster, call="action")


def test_list_clusters_by_datacenter_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_clusters_by_datacenter
    with anything other than --function or -f.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.list_clusters_by_datacenter, call="action"
    )


def test_list_hosts_by_datacenter_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_hosts_by_datacenter
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_hosts_by_datacenter, call="action")


def test_list_hbas_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_hbas
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_hbas, call="action")


def test_list_dvs_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_dvs
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_dvs, call="action")


def test_list_vapps_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_vapps
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_vapps, call="action")


def test_list_templates_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_templates
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.list_templates, call="action")


def test_create_datacenter_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call create_datacenter
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.create_datacenter, call="action")


def test_create_cluster_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call create_cluster
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.create_cluster, call="action")


def test_rescan_hba_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call rescan_hba
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.rescan_hba, call="action")


def test_upgrade_tools_all_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call upgrade_tools_all
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.upgrade_tools_all, call="action")


def test_enter_maintenance_mode_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call enter_maintenance_mode
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.enter_maintenance_mode, call="action")


def test_exit_maintenance_mode_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call exit_maintenance_mode
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.exit_maintenance_mode, call="action")


def test_create_folder_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call create_folder
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.create_folder, call="action")


def test_add_host_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call add_host
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.add_host, call="action")


def test_remove_host_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call remove_host
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.remove_host, call="action")


def test_connect_host_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call connect_host
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.connect_host, call="action")


def test_disconnect_host_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call disconnect_host
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.disconnect_host, call="action")


def test_reboot_host_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call reboot_host
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.reboot_host, call="action")


def test_create_datastore_cluster_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call create_datastore_cluster
    with anything other than --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.create_datastore_cluster, call="action")


def test_show_instance_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call show_instance
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.show_instance, name=vm_name, call="function"
    )


def test_start_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call start
    with anything other than --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.start, name=vm_name, call="function")


def test_stop_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call stop
    with anything other than --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.stop, name=vm_name, call="function")


def test_suspend_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call suspend
    with anything other than --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.suspend, name=vm_name, call="function")


def test_reset_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call reset
    with anything other than --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.reset, name=vm_name, call="function")


def test_terminate_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call terminate
    with anything other than --action or -a.
    """
    pytest.raises(SaltCloudSystemExit, vmware.terminate, name=vm_name, call="function")


def test_destroy_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call destroy
    with --function or -f.
    """
    pytest.raises(SaltCloudSystemExit, vmware.destroy, name=vm_name, call="function")


def test_shutdown_host_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call convert_to_template
    with anything other than --action or -a.
    """
    with patch.object(vmware, "_get_si", Mock()), patch(
        "salt.utils.vmware.get_mor_by_property", Mock()
    ):
        pytest.raises(
            SaltCloudSystemExit,
            vmware.shutdown_host,
            kwargs={"host": vm_name},
            call="action",
        )


def test_upgrade_tools_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call upgrade_tools
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.upgrade_tools, name=vm_name, call="function"
    )


def test_create_snapshot_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call create_snapshot
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.create_snapshot, name=vm_name, call="function"
    )


def test_revert_to_snapshot_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call revert_to_snapshot
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.revert_to_snapshot,
        name=vm_name,
        call="function",
    )


def test_remove_snapshot_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call remove_snapshot
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.remove_snapshot,
        name=vm_name,
        kwargs={"snapshot_name": "mySnapshot"},
        call="function",
    )


def test_remove_snapshot_call_no_snapshot_name_in_kwargs(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when name is not present in kwargs.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.remove_snapshot, name=vm_name, call="action"
    )


def test_remove_all_snapshots_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call remove_all_snapshots
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.remove_all_snapshots,
        name=vm_name,
        call="function",
    )


def test_convert_to_template_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call convert_to_template
    with anything other than --action or -a.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.convert_to_template,
        name=vm_name,
        call="function",
    )


def test_avail_sizes():
    """
    Tests that avail_sizes returns an empty dictionary.
    """
    assert vmware.avail_sizes(call="foo") == {}


def test_create_datacenter_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    create_datacenter.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.create_datacenter, kwargs=None, call="function"
    )


def test_create_datacenter_no_name_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when name is not present in
    kwargs that are provided to create_datacenter.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datacenter,
        kwargs={"foo": "bar"},
        call="function",
    )


def test_create_datacenter_name_too_short():
    """
    Tests that a SaltCloudSystemExit is raised when name is present in kwargs
    that are provided to create_datacenter but is an empty string.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datacenter,
        kwargs={"name": ""},
        call="function",
    )


def test_create_datacenter_name_too_long():
    """
    Tests that a SaltCloudSystemExit is raised when name is present in kwargs
    that are provided to create_datacenter but is a string with length <= 80.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datacenter,
        kwargs={
            "name": "cCD2GgJGPG1DUnPeFBoPeqtdmUxIWxDoVFbA14vIG0BPoUECkgbRMnnY6gaUPBvIDCcsZ5HU48ubgQu5c"
        },
        call="function",
    )


def test_create_cluster_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    create_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.create_cluster, kwargs=None, call="function"
    )


def test_create_cluster_no_name_no_datacenter_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when neither the name nor the
    datacenter is present in kwargs that are provided to create_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_cluster,
        kwargs={"foo": "bar"},
        call="function",
    )


def test_create_cluster_no_datacenter_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when the name is present but the
    datacenter is not present in kwargs that are provided to create_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_cluster,
        kwargs={"name": "my-cluster"},
        call="function",
    )


def test_create_cluster_no_name_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when the datacenter is present
    but the name is not present in kwargs that are provided to create_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_cluster,
        kwargs={"datacenter": "my-datacenter"},
        call="function",
    )


def test_rescan_hba_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    rescan_hba.
    """
    pytest.raises(SaltCloudSystemExit, vmware.rescan_hba, kwargs=None, call="function")


def test_rescan_hba_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to rescan_hba.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.rescan_hba,
        kwargs={"foo": "bar"},
        call="function",
    )


def test_create_snapshot_no_kwargs(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    create_snapshot.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_snapshot,
        name=vm_name,
        kwargs=None,
        call="action",
    )


def test_create_snapshot_no_snapshot_name_in_kwargs(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when snapshot_name is not present
    in kwargs that are provided to create_snapshot.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_snapshot,
        name=vm_name,
        kwargs={"foo": "bar"},
        call="action",
    )


def test_add_host_no_esxi_host_user_in_config():
    """
    Tests that a SaltCloudSystemExit is raised when esxi_host_user is not
    specified in the cloud provider configuration when calling add_host.
    """
    with pytest.raises(
        SaltCloudSystemExit,
        match="You must specify the ESXi host username in your providers config.",
    ):
        vmware.add_host(kwargs=None, call="function")


def test_add_host_no_esxi_host_password_in_config():
    """
    Tests that a SaltCloudSystemExit is raised when esxi_host_password is not
    specified in the cloud provider configuration when calling add_host.
    """
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"],
        {"esxi_host_user": "root"},
        clean=True,
    ):
        with pytest.raises(
            SaltCloudSystemExit,
            match="You must specify the ESXi host password in your providers config.",
        ):
            vmware.add_host(kwargs=None, call="function")


def test_no_clonefrom_just_image(profile):
    """
    Tests that the profile is configured correctly when deploying using an image
    """

    profile_additions = {"image": "some-image.iso"}
    vm_ = {"profile": profile}
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"]["profiles"]["base-gold"],
        profile_additions,
        clean=True,
    ):
        assert (
            config.is_profile_configured(
                vmware.__opts__, "vcenter01:vmware", "base-gold", vm_=vm_
            )
            is True
        )


def test_just_clonefrom(profile):
    """
    Tests that the profile is configured correctly when deploying by cloning from a template
    """

    profile_additions = {
        "clonefrom": "test-template",
        "image": "should ignore image",
    }
    vm_ = {"profile": profile}
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"]["profiles"]["base-gold"],
        profile_additions,
        clean=True,
    ):
        assert (
            config.is_profile_configured(
                vmware.__opts__, "vcenter01:vmware", "base-gold", vm_=vm_
            )
            is True
        )


def test_just_Instantclonefrom(vm_name):
    """
    Tests that the profile is configured correctly when deploying by instant cloning from a running VM
    """

    profile_additions = {
        "clonefrom": vm_name,
        "instant_clone": True,
    }
    vm_ = {"profile": profile}
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"]["profiles"]["base-gold"],
        profile_additions,
        clean=True,
    ):
        assert (
            config.is_profile_configured(
                vmware.__opts__, "vcenter01:vmware", "base-gold", vm_=vm_
            )
            is True
        )


def test_add_new_ide_controller_helper():
    """
    Tests that creating a new controller, ensuring that it will generate a controller key
    if one is not provided
    """
    with patch("salt.cloud.clouds.vmware.randint", return_value=101) as randint_mock:
        controller_label = "Some label"
        bus_number = 1
        spec = vmware._add_new_ide_controller_helper(controller_label, None, bus_number)
        assert spec.device.key == randint_mock.return_value

        spec = vmware._add_new_ide_controller_helper(controller_label, 200, bus_number)
        assert spec.device.key == 200

        assert spec.device.busNumber == bus_number
        assert spec.device.deviceInfo.label == controller_label
        assert spec.device.deviceInfo.summary == controller_label


def test_manage_devices_just_cd():
    """
    Tests that when adding IDE/CD drives, controller keys will be in the apparent
    safe-range on ESX 5.5 but randomly generated on other versions (i.e. 6)
    """
    device_map = {
        "ide": {"IDE 0": {}, "IDE 1": {}},
        "cd": {"CD/DVD Drive 1": {"controller": "IDE 0"}},
    }
    with patch(
        "salt.cloud.clouds.vmware.get_vcenter_version",
        return_value="VMware ESXi 5.5.0",
    ):
        specs = vmware._manage_devices(device_map, vm=None)["device_specs"]

        assert specs[0].device.key == vmware.SAFE_ESX_5_5_CONTROLLER_KEY_INDEX
        assert specs[1].device.key == vmware.SAFE_ESX_5_5_CONTROLLER_KEY_INDEX + 1
        assert specs[2].device.controllerKey == vmware.SAFE_ESX_5_5_CONTROLLER_KEY_INDEX

    with patch(
        "salt.cloud.clouds.vmware.get_vcenter_version", return_value="VMware ESXi 6"
    ):
        with patch("salt.cloud.clouds.vmware.randint", return_value=100) as first_key:
            specs = vmware._manage_devices(device_map, vm=None)["device_specs"]

            assert specs[0].device.key == first_key.return_value
            assert specs[2].device.controllerKey == first_key.return_value


def test_add_host_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to add_host.
    """
    provider_config_additions = {
        "esxi_host_user": "root",
        "esxi_host_password": "myhostpassword",
    }
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"],
        provider_config_additions,
        clean=True,
    ):
        with pytest.raises(
            SaltCloudSystemExit,
            match="You must specify either the IP or DNS name of the host system.",
        ):
            vmware.add_host(kwargs={"foo": "bar"}, call="function")


def test_add_host_both_cluster_and_datacenter_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when both cluster and datacenter
    are present in kwargs that are provided to add_host.
    """
    provider_config_additions = {
        "esxi_host_user": "root",
        "esxi_host_password": "myhostpassword",
    }
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"],
        provider_config_additions,
        clean=True,
    ):
        with pytest.raises(
            SaltCloudSystemExit,
            match="You must specify either the cluster name or the datacenter name.",
        ):
            vmware.add_host(
                kwargs={
                    "host": "my-esxi-host",
                    "datacenter": "my-datacenter",
                    "cluster": "my-cluster",
                },
                call="function",
            )


def test_add_host_neither_cluster_nor_datacenter_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when neither cluster nor
    datacenter is present in kwargs that are provided to add_host.
    """
    provider_config_additions = {
        "esxi_host_user": "root",
        "esxi_host_password": "myhostpassword",
    }
    with patch.dict(
        vmware.__opts__["providers"]["vcenter01"]["vmware"],
        provider_config_additions,
        clean=True,
    ):
        with pytest.raises(
            SaltCloudSystemExit,
            match="You must specify either the cluster name or the datacenter name.",
        ):
            vmware.add_host(kwargs={"host": "my-esxi-host"}, call="function")


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_add_host_cluster_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified cluster present
    in kwargs that are provided to add_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            provider_config_additions = {
                "esxi_host_user": "root",
                "esxi_host_password": "myhostpassword",
            }

            with patch.dict(
                vmware.__opts__["providers"]["vcenter01"]["vmware"],
                provider_config_additions,
                clean=True,
            ):
                with pytest.raises(
                    SaltCloudSystemExit, match="Specified cluster does not exist."
                ):
                    vmware.add_host(
                        kwargs={"host": "my-esxi-host", "cluster": "my-cluster"},
                        call="function",
                    )


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_add_host_datacenter_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified datacenter
    present in kwargs that are provided to add_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            provider_config_additions = {
                "esxi_host_user": "root",
                "esxi_host_password": "myhostpassword",
            }
            with patch.dict(
                vmware.__opts__["providers"]["vcenter01"]["vmware"],
                provider_config_additions,
                clean=True,
            ):
                with pytest.raises(
                    SaltCloudSystemExit, match="Specified datacenter does not exist."
                ):
                    vmware.add_host(
                        kwargs={"host": "my-esxi-host", "datacenter": "my-datacenter"},
                        call="function",
                    )


def test_remove_host_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    remove_host.
    """
    pytest.raises(SaltCloudSystemExit, vmware.remove_host, kwargs=None, call="function")


def test_remove_host_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to remove_host.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.remove_host,
        kwargs={"foo": "bar"},
        call="function",
    )


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_remove_host_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified host present
    in kwargs that are provided to remove_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            pytest.raises(
                SaltCloudSystemExit,
                vmware.remove_host,
                kwargs={"host": "my-host"},
                call="function",
            )


def test_connect_host_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    connect_host.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.connect_host, kwargs=None, call="function"
    )


def test_connect_host_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to connect_host.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.connect_host,
        kwargs={"foo": "bar"},
        call="function",
    )


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_connect_host_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified host present
    in kwargs that are provided to connect_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            pytest.raises(
                SaltCloudSystemExit,
                vmware.connect_host,
                kwargs={"host": "my-host"},
                call="function",
            )


def test_disconnect_host_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    disconnect_host.
    """
    pytest.raises(
        SaltCloudSystemExit, vmware.disconnect_host, kwargs=None, call="function"
    )


def test_disconnect_host_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to disconnect_host.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.disconnect_host,
        kwargs={"foo": "bar"},
        call="function",
    )


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_disconnect_host_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified host present
    in kwargs that are provided to disconnect_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            pytest.raises(
                SaltCloudSystemExit,
                vmware.disconnect_host,
                kwargs={"host": "my-host"},
                call="function",
            )


def test_reboot_host_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    reboot_host.
    """
    pytest.raises(SaltCloudSystemExit, vmware.reboot_host, kwargs=None, call="function")


def test_reboot_host_no_host_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when host is not present in
    kwargs that are provided to reboot_host.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.reboot_host,
        kwargs={"foo": "bar"},
        call="function",
    )


@pytest.mark.skipif(
    HAS_LIBS is False, reason="Install pyVmomi to be able to run this unit test."
)
def test_reboot_host_not_exists():
    """
    Tests that a SaltCloudSystemExit is raised when the specified host present
    in kwargs that are provided to connect_host does not exist in the VMware
    environment.
    """
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_by_property", MagicMock(return_value=None)
        ):
            pytest.raises(
                SaltCloudSystemExit,
                vmware.reboot_host,
                kwargs={"host": "my-host"},
                call="function",
            )


def test_create_datastore_cluster_no_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when no kwargs are provided to
    create_datastore_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datastore_cluster,
        kwargs=None,
        call="function",
    )


def test_create_datastore_cluster_no_name_in_kwargs():
    """
    Tests that a SaltCloudSystemExit is raised when name is not present in
    kwargs that are provided to create_datastore_cluster.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datastore_cluster,
        kwargs={"foo": "bar"},
        call="function",
    )


def test_create_datastore_cluster_name_too_short():
    """
    Tests that a SaltCloudSystemExit is raised when name is present in kwargs
    that are provided to create_datastore_cluster but is an empty string.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datastore_cluster,
        kwargs={"name": ""},
        call="function",
    )


def test_create_datastore_cluster_name_too_long():
    """
    Tests that a SaltCloudSystemExit is raised when name is present in kwargs
    that are provided to create_datastore_cluster but is a string with length <= 80.
    """
    pytest.raises(
        SaltCloudSystemExit,
        vmware.create_datastore_cluster,
        kwargs={
            "name": "cCD2GgJGPG1DUnPeFBoPeqtdmUxIWxDoVFbA14vIG0BPoUECkgbRMnnY6gaUPBvIDCcsZ5HU48ubgQu5c"
        },
        call="function",
    )


def test__add_new_hard_disk_helper():
    with patch("salt.cloud.clouds.vmware._get_si", MagicMock(return_value=None)):
        with patch(
            "salt.utils.vmware.get_mor_using_container_view",
            side_effect=[None, None],
        ):
            pytest.raises(
                SaltCloudSystemExit,
                vmware._add_new_hard_disk_helper,
                disk_label="test",
                size_gb=100,
                unit_number=0,
                datastore="whatever",
            )
        with patch(
            "salt.utils.vmware.get_mor_using_container_view",
            side_effect=["Datastore", None],
        ):
            pytest.raises(
                AttributeError,
                vmware._add_new_hard_disk_helper,
                disk_label="test",
                size_gb=100,
                unit_number=0,
                datastore="whatever",
            )
            vmware.salt.utils.vmware.get_mor_using_container_view.assert_called_with(
                None, vim.Datastore, "whatever"
            )
        with patch(
            "salt.utils.vmware.get_mor_using_container_view",
            side_effect=[None, "Cluster"],
        ):
            pytest.raises(
                AttributeError,
                vmware._add_new_hard_disk_helper,
                disk_label="test",
                size_gb=100,
                unit_number=0,
                datastore="whatever",
            )
            vmware.salt.utils.vmware.get_mor_using_container_view.assert_called_with(
                None, vim.StoragePod, "whatever"
            )



"""
    Log4Mongo Logging Handler
    =========================

    This module provides a logging handler for sending salt logs to MongoDB

    Configuration
    -------------

    In the salt configuration file (e.g. /etc/salt/{master,minion}):

    .. code-block:: yaml

        log4mongo_handler:
          host: mongodb_host
          port: 27017
          database_name: logs
          collection: salt_logs
          username: logging
          password: reindeerflotilla
          write_concern: 0
          log_level: warning


    Log Level
    .........

    If not set, the log_level will be set to the level defined in the global
    configuration file setting.

    .. admonition:: Inspiration

        This work was inspired by the Salt logging handlers for LogStash and
        Sentry and by the log4mongo Python implementation.
"""

import logging
import socket

from salt._logging import LOG_LEVELS

try:
    from log4mongo.handlers import MongoFormatter, MongoHandler

    HAS_MONGO = True
except ImportError:
    HAS_MONGO = False

__virtualname__ = "mongo"


def __virtual__():
    if not HAS_MONGO:
        return False
    return __virtualname__


class FormatterWithHost(logging.Formatter):
    def format(self, record):
        mongoformatter = MongoFormatter()
        document = mongoformatter.format(record)
        document["hostname"] = socket.gethostname()
        return document


def setup_handlers():
    handler_id = "log4mongo_handler"
    if handler_id in __opts__:
        config_fields = {
            "host": "host",
            "port": "port",
            "database_name": "database_name",
            "collection": "collection",
            "username": "username",
            "password": "password",
            "write_concern": "w",
        }

        config_opts = {}
        for config_opt, arg_name in config_fields.items():
            config_opts[arg_name] = __opts__[handler_id].get(config_opt)

        config_opts["level"] = LOG_LEVELS[
            __opts__[handler_id].get("log_level", __opts__.get("log_level", "error"))
        ]

        handler = MongoHandler(formatter=FormatterWithHost(), **config_opts)
        yield handler
    else:
        yield False



import pytest

pytestmark = [
    pytest.mark.skip_unless_on_linux,
    pytest.mark.skipif(
        'grains["os_family"] != "Debian"',
        reason="Tests applicable only to Debian and Ubuntu",
    ),
]


def test_grains(grains):
    """
    Test to ensure that the lsb_distrib_xxxx grains are
    populated on Debian machines
    """
    assert "lsb_distrib_id" in grains
    assert "lsb_distrib_release" in grains
    assert "lsb_distrib_codename" in grains

    assert grains["lsb_distrib_id"] == grains["osfullname"]
    assert grains["lsb_distrib_release"] == grains["osrelease"]
    assert grains["lsb_distrib_codename"] == grains["oscodename"]
